/*---------------------------------------------------------------------------*/
/* Program:  align.c                                                         */
/*                                                                           */
/* Purpose:  This program is specifically designed to align the cell         */
/*           images obtained from UMKC which are 512x480 with four           */
/*           subimages which are 256x240, slightly out of alignment.         */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     February 11, 1994                                               */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Global variables */
FLOAT_TYPE **Data1;
FLOAT_TYPE **Data2;
int Xdim, Ydim, Zdim;
int bestx[3], besty[3];
int Debug = FALSE;
int DoAlign = TRUE;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine aligns patches in A to B.                          */
/*---------------------------------------------------------------------------*/
void MatchAB(int x1, int y1, int x2, int y2, int *best_dx, int *best_dy)
{
   int x, y, dx, dy, range = 10;
   float diff, error, best_err;

   /* Search for best dx,dy alignment of patches */
   best_err = (float) (x2 - x1) * (y2 - y1) * 255;
   *best_dx = 0;
   *best_dy = 0;
   for (dx = -range; dx <= range; dx++)
      for (dy = -range; dy <= range; dy++)
      {
	 error = 0.0;
	 for (y = y1; y < y2; y++)
	    for (x = x1; x < x2; x++)
	    {
	       diff = Data1[y + dy][x + dx] - Data1[y][x + Xdim / 2];
	       error += diff * diff;
	    }
	 if (error < best_err)
	 {
	    best_err = error;
	    *best_dx = dx;
	    *best_dy = dy;
	 }
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine aligns patches in A to C.                          */
/*---------------------------------------------------------------------------*/
void MatchAC(int x1, int y1, int x2, int y2, int *best_dx, int *best_dy)
{
   int x, y, dx, dy, range = 10;
   float diff, error, best_err;

   /* Search for best dx,dy alignment of patches */
   best_err = (float) (x2 - x1) * (y2 - y1) * 255;
   *best_dx = 0;
   *best_dy = 0;
   for (dx = -range; dx <= range; dx++)
      for (dy = -range; dy <= range; dy++)
      {
	 error = 0.0;
	 for (y = y1; y < y2; y++)
	    for (x = x1; x < x2; x++)
	    {
	       diff = Data1[y + dy][x + dx] - Data1[y + Ydim / 2][x];
	       error += diff * diff;
	    }
	 if (error < best_err)
	 {
	    best_err = error;
	    *best_dx = dx;
	    *best_dy = dy;
	 }
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine aligns patches in A to D.                          */
/*---------------------------------------------------------------------------*/
void MatchAD(int x1, int y1, int x2, int y2, int *best_dx, int *best_dy)
{
   int x, y, dx, dy, range = 10;
   float diff, error, best_err;

   /* Search for best dx,dy alignment of patches */
   best_err = (float) (x2 - x1) * (y2 - y1) * 255;
   *best_dx = 0;
   *best_dy = 0;
   for (dx = -range; dx <= range; dx++)
      for (dy = -range; dy <= range; dy++)
      {
	 error = 0.0;
	 for (y = y1; y < y2; y++)
	    for (x = x1; x < x2; x++)
	    {
	       diff = Data1[y + dy][x + dx] - Data1[y + Ydim / 2][x + Xdim / 2];
	       error += diff * diff;
	    }
	 if (error < best_err)
	 {
	    best_err = error;
	    *best_dx = dx;
	    *best_dy = dy;
	 }
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs subimage alignment.                       */
/*---------------------------------------------------------------------------*/
void Align()
{
   int x, y, newx, newy;
   int X1, X2, X3, X4, X5, Y1, Y2, Y3, Y4, Y5;
   int fringe = 30;

   /* Set range of search windows */
   X1 = fringe;
   X2 = Xdim / 8 + fringe / 2;
   X3 = Xdim / 4;
   X4 = 3 * Xdim / 8 - fringe / 2;
   X5 = Xdim / 2 - fringe;
   Y1 = fringe;
   Y2 = Ydim / 8 + fringe / 2;
   Y3 = Ydim / 4;
   Y4 = 3 * Ydim / 8 - fringe / 2;
   Y5 = Ydim / 2 - fringe;

   /* Search for automatic alignment offsets */
   if (DoAlign == TRUE)
   {
      if (Debug == TRUE)
      {
	 MatchAB(X1, Y1, X3, Y3, &(bestx[0]), &(besty[0]));
	 printf("align A to B UL with shift %d %d\n", bestx[0], besty[0]);

	 MatchAB(X2, Y1, X4, Y3, &(bestx[0]), &(besty[0]));
	 printf("align A to B UC with shift %d %d\n", bestx[0], besty[0]);

	 MatchAB(X3, Y1, X5, Y3, &(bestx[0]), &(besty[0]));
	 printf("align A to B UR with shift %d %d\n", bestx[0], besty[0]);

	 MatchAB(X1, Y2, X3, Y4, &(bestx[0]), &(besty[0]));
	 printf("align A to B CL with shift %d %d\n", bestx[0], besty[0]);

	 MatchAB(X2, Y2, X4, Y4, &(bestx[0]), &(besty[0]));
	 printf("align A to B CC with shift %d %d\n", bestx[0], besty[0]);

	 MatchAB(X3, Y2, X5, Y4, &(bestx[0]), &(besty[0]));
	 printf("align A to B CR with shift %d %d\n", bestx[0], besty[0]);

	 MatchAB(X1, Y3, X3, Y5, &(bestx[0]), &(besty[0]));
	 printf("align A to B LL with shift %d %d\n", bestx[0], besty[0]);

	 MatchAB(X2, Y3, X4, Y5, &(bestx[0]), &(besty[0]));
	 printf("align A to B LC with shift %d %d\n", bestx[0], besty[0]);

	 MatchAB(X3, Y3, X5, Y5, &(bestx[0]), &(besty[0]));
	 printf("align A to B LR with shift %d %d\n", bestx[0], besty[0]);
      }

      /* Align A to B */
      MatchAB(X1, Y1, X5, Y5, &(bestx[0]), &(besty[0]));
      printf("align A to B with shift %d %d\n\n", bestx[0], besty[0]);

      if (Debug == TRUE)
      {
	 MatchAC(X1, Y1, X3, Y3, &(bestx[1]), &(besty[1]));
	 printf("align A to C UL with shift %d %d\n", bestx[1], besty[1]);

	 MatchAC(X2, Y1, X4, Y3, &(bestx[1]), &(besty[1]));
	 printf("align A to C UC with shift %d %d\n", bestx[1], besty[1]);

	 MatchAC(X3, Y1, X5, Y3, &(bestx[1]), &(besty[1]));
	 printf("align A to C UR with shift %d %d\n", bestx[1], besty[1]);

	 MatchAC(X1, Y2, X3, Y4, &(bestx[1]), &(besty[1]));
	 printf("align A to C CL with shift %d %d\n", bestx[1], besty[1]);

	 MatchAC(X2, Y2, X4, Y4, &(bestx[1]), &(besty[1]));
	 printf("align A to C CC with shift %d %d\n", bestx[1], besty[1]);

	 MatchAC(X3, Y2, X5, Y4, &(bestx[1]), &(besty[1]));
	 printf("align A to C CR with shift %d %d\n", bestx[1], besty[1]);

	 MatchAC(X1, Y3, X3, Y5, &(bestx[1]), &(besty[1]));
	 printf("align A to C LL with shift %d %d\n", bestx[1], besty[1]);

	 MatchAC(X2, Y3, X4, Y5, &(bestx[1]), &(besty[1]));
	 printf("align A to C LC with shift %d %d\n", bestx[1], besty[1]);

	 MatchAC(X3, Y3, X5, Y5, &(bestx[1]), &(besty[1]));
	 printf("align A to C LR with shift %d %d\n", bestx[1], besty[1]);
      }

      /* Align A to C */
      MatchAC(X1, Y1, X5, Y5, &(bestx[1]), &(besty[1]));
      printf("align A to C with shift %d %d\n\n", bestx[1], besty[1]);

      if (Debug == TRUE)
      {
	 MatchAD(X1, Y1, X3, Y3, &(bestx[2]), &(besty[2]));
	 printf("align A to D UL with shift %d %d\n", bestx[2], besty[2]);

	 MatchAD(X2, Y1, X4, Y3, &(bestx[2]), &(besty[2]));
	 printf("align A to D UC with shift %d %d\n", bestx[2], besty[2]);

	 MatchAD(X3, Y1, X5, Y3, &(bestx[2]), &(besty[2]));
	 printf("align A to D UR with shift %d %d\n", bestx[2], besty[2]);

	 MatchAD(X1, Y2, X3, Y4, &(bestx[2]), &(besty[2]));
	 printf("align A to D CL with shift %d %d\n", bestx[2], besty[2]);

	 MatchAD(X2, Y2, X4, Y4, &(bestx[2]), &(besty[2]));
	 printf("align A to D CC with shift %d %d\n", bestx[2], besty[2]);

	 MatchAD(X3, Y2, X5, Y4, &(bestx[2]), &(besty[2]));
	 printf("align A to D CR with shift %d %d\n", bestx[2], besty[2]);

	 MatchAD(X1, Y3, X3, Y5, &(bestx[2]), &(besty[2]));
	 printf("align A to D LL with shift %d %d\n", bestx[2], besty[2]);

	 MatchAD(X2, Y3, X4, Y5, &(bestx[2]), &(besty[2]));
	 printf("align A to D LC with shift %d %d\n", bestx[2], besty[2]);

	 MatchAD(X3, Y3, X5, Y5, &(bestx[2]), &(besty[2]));
	 printf("align A to D LR with shift %d %d\n", bestx[2], besty[2]);
      }

      /* Align A to D */
      MatchAD(X1, Y1, X5, Y5, &(bestx[2]), &(besty[2]));
      printf("align A to D with shift %d %d\n\n", bestx[2], besty[2]);
   }

   /* Copy A data */
   for (y = 0; y < (Ydim / 2); y++)
      for (x = 0; x < (Xdim / 2); x++)
	 Data2[y][x] = Data1[y][x];

   /* Copy B data */
   for (y = 0; y < (Ydim / 2); y++)
      for (x = (Xdim / 2); x < Xdim; x++)
      {
	 newx = x - bestx[0];
	 newy = y - besty[0];
	 if ((newy < 0) || (newy >= (Ydim / 2)) ||
	     (newx < (Xdim / 2)) || (newx >= Xdim))
	    Data2[y][x] = Data1[y][x];
	 else
	    Data2[y][x] = Data1[y - besty[0]][x - bestx[0]];
      }

   /* Copy C data */
   for (y = (Ydim / 2); y < Ydim; y++)
      for (x = 0; x < (Xdim / 2); x++)
      {
	 newx = x - bestx[1];
	 newy = y - besty[1];
	 if ((newx < 0) || (newx >= (Xdim / 2)) ||
	     (newy < (Ydim / 2)) || (newy >= Ydim))
	    Data2[y][x] = Data1[y][x];
	 else
	    Data2[y][x] = Data1[y - besty[1]][x - bestx[1]];
      }

   /* Copy D data */
   for (y = (Ydim / 2); y < Ydim; y++)
      for (x = (Xdim / 2); x < Xdim; x++)
      {
	 newx = x - bestx[2];
	 newy = y - besty[2];
	 if ((newx < (Xdim / 2)) || (newx >= Xdim) ||
	     (newy < (Ydim / 2)) || (newy >= Ydim))
	    Data2[y][x] = Data1[y][x];
	 else
	    Data2[y][x] = Data1[y - besty[2]][x - bestx[2]];
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   int PixType, DimCnt;

   /* Program variables */
   int i = 0, j;

   /* Interpret program options */
   printf("ALIGN Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'a':
	 for (j = 0; j < 3; j++)
	 {
	    if (sscanf(argv[++i], "%d", &(bestx[j])) == 0)
	       Error("Could not get bestx argument");
	    if (sscanf(argv[++i], "%d", &(besty[j])) == 0)
	       Error("Could not get besty argument");
	 }
	 DoAlign = FALSE;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: align [options] infile outfile\n");
      fprintf(stderr, "       [-d]  Print debugging information\n");
      fprintf(stderr, "       [-a # # # # # #]  Alignment offsets\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);

   /* Align the four subimages */
   Align();

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   return (0);
}
