/*---------------------------------------------------------------------------*/
/* Program:  back_project.c                                                  */
/*                                                                           */
/* Purpose:  This program take the filtered image projections from           */
/*           PROJECT and FILTER_PROJECT and reconstructs the original        */
/*           image using parallel back projection.                           */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     May 11, 1994                                                    */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define ROUND(r) ((r) >= 0.0) ? ((int)(r + 0.5)) : ((int)(r - 0.5))

/* Global variables */
int Debug = FALSE;
int Xdim, Ydim, Zdim;
int NumBuckets, NumProject;
FLOAT_TYPE **Data1;
FLOAT_TYPE **Data2;
double sin();
double cos();

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs back projection.                          */
/*---------------------------------------------------------------------------*/
void BackProject()
{
   int x, y, a, b, Xcenter, Ycenter;
   float val, Position, Scale;
   double *SinTab, *CosTab;

   /* Initialize sin and cos lookup tables */
   SinTab = (double *) malloc((unsigned) sizeof(double) * NumProject);
   CosTab = (double *) malloc((unsigned) sizeof(double) * NumProject);
   for (a = 0; a < NumProject; a++)
   {
      SinTab[a] = sin((double) a * M_PI / NumProject);
      CosTab[a] = cos((double) a * M_PI / NumProject);
   }

   /* Initialize output image to zero */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 Data2[y][x] = 0;

   /* Project each pixel in the image */
   Scale = (float) (1.0 / (NumProject * NumProject));
   Xcenter = Xdim / 2;
   Ycenter = Ydim / 2;
   for (y = -Ycenter; y < Ycenter; y++)
      for (x = -Xcenter; x < Xcenter; x++)
      {
	 val = 0.0;
	 for (a = 0; a < NumProject; a++)
	 {
	    Position = (float) (x * CosTab[a] + y * SinTab[a]);
	    b = ROUND(Position);
	    if (b < 0)
	       b += NumBuckets;
	    val += Data1[a][b];
	 }
	 Data2[y + Ycenter][x + Xcenter] = val * Scale;
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
void main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   int PixType, DimCnt;

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("BACK_PROJECT Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: back_project [options] infile outfile\n");
      fprintf(stderr, "       [-d]  Print debugging information\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &NumBuckets, &NumProject, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Xdim = Ydim = NumBuckets / 2;
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   BackProject();

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   exit(0);
}
