/*---------------------------------------------------------------------------*/
/* Program:  blob_extract.c                                                  */
/*                                                                           */
/* Purpose:  This program scans a b/w image in KUIM format and generates     */
/*           a collection of b/w output images of a fixed size which         */
/*           contain one blob each.  This can be used to extract characters  */
/*           from images to construct a font database.                       */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     April 15, 1997                                                  */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994-1997               */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Global program variables */
SHORT_TYPE **data1;
SHORT_TYPE **data2;
int Xdim, Ydim, Zdim;
int NewXdim = 100, NewYdim = 100, NewZdim = 1;
int Color = 2;
int Size = 0;
int MeanX = 0;
int MeanY = 0;
int MeanZ = 0;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine marks 2D blobs.                                    */
/*---------------------------------------------------------------------------*/
void BlobColor(int x, int y, int c)
{
   if (data1[y][x] == 1)
   {
      data1[y][x] = c;
      MeanX += x;
      MeanY += y;
      Size++;
      if (x + 1 < Xdim)
	 BlobColor(x + 1, y, c);
      if (x > 0)
	 BlobColor(x - 1, y, c);
      if (y + 1 < Ydim)
	 BlobColor(x, y + 1, c);
      if (y > 0)
	 BlobColor(x, y - 1, c);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine copies 2D blobs.                                   */
/*---------------------------------------------------------------------------*/
void BlobCopy(int x, int y, int c)
{
   int x2, y2;

   if (data1[y][x] == c)
   {
      data1[y][x] = 0;
      y2 = y - MeanY + 50;
      x2 = x - MeanX + 50;
      if ((y2 > 0) && (y2 < NewYdim) && (x2 > 0) && (x2 < NewXdim))
	 data2[y2][x2] = 1;
      if (x + 1 < Xdim)
	 BlobCopy(x + 1, y, c);
      if (x > 0)
	 BlobCopy(x - 1, y, c);
      if (y + 1 < Ydim)
	 BlobCopy(x, y + 1, c);
      if (y > 0)
	 BlobCopy(x, y - 1, c);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   int PixType, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int Reverse = FALSE;
   int MinSize = 100;
   int MaxSize = 2000;
   int i = 0, x, y;

   /* Interpret program options */
   printf("BLOB_EXTRACT Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'o':
	 if (sscanf(argv[++i], "%d %d", &NewXdim, &NewYdim) == 0)
	    Error("Could not get integer argument");
	 break;
      case 't':
	 if (sscanf(argv[++i], "%d", &MinSize) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'T':
	 if (sscanf(argv[++i], "%d", &MaxSize) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'r':
	 Reverse = TRUE;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: blob_extract [options] infile outfile\n");
      fprintf(stderr, "       [-d]      Print debugging information\n");
      fprintf(stderr, "       [-r]      Reverse 0/1 before blob coloring\n");
      fprintf(stderr, "       [-t #]    Minimum blob size (10)\n");
      fprintf(stderr, "       [-T #]    Maximum blob size (2000)\n");
      fprintf(stderr, "       [-o # #]  Output images size (100 100)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Open input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can only process 2D images");

   /* Read input image */
   data1 = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);
   im_read(Image1, SHORT, (char *) &(data1[0][0]));

   /* Preprocess the image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 if (data1[y][x] > 1)
	    data1[y][x] = 1;
	 if (data1[y][x] < 0)
	    data1[y][x] = 0;
	 if (Reverse == TRUE)
	    data1[y][x] = 1 - data1[y][x];
      }

   /* Find the blobs in the input image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 if (data1[y][x] == 1)
	 {
	    MeanX = MeanY = 0;
	    Size = 0;
	    BlobColor(x, y, Color);
	    if ((Size >= MinSize) && (Size <= MaxSize))
	    {
	       /* Calculate blob center */
	       MeanX /= Size;
	       MeanY /= Size;
	       if (Debug == TRUE)
		  printf("blob %4d size %4d pos %5d %5d\n",
		     Color, Size, (int) (0.5 + MeanX), (int) (0.5 + MeanY));

	       /* Create output image */
	       sprintf(Name3, "%s%02d", Name2, Color);
	       Image2 = im_create(Name3, SHORT, NewXdim, NewYdim, NewZdim);
	       data2 = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);

	       /* Write information to output image */
	       BlobCopy(x, y, Color);
	       im_write(Image2, SHORT, (char *) &(data2[0][0]));
	       im_free2D((char **) data2);
	       Color++;
	    }
	 }
   return (0);
}
