/*--------------------------------------------------------------------------*/
/* Program:  color_to.c                                                     */
/*                                                                          */
/* Purpose:  This program converts RGB color images to 3D multichannel      */
/*           images in KUIM format.  The type of color conversion needed    */
/*           is specified by input switches.                                */
/*                                                                          */
/* Author:   John Gauch, Xiaodong Hong                                      */
/*                                                                          */
/* Date:     September 26, 1996   - Original program.                         */
/*           November 1 1996    - Add uvw xyz color systems.                */
/*           April 8, 1997      - Add HIS, U*V*W* and Lab color systems.    */
/*                                                                          */
/* Note:     Copyright (C) The University of Kansas, 1994                   */
/*--------------------------------------------------------------------------*/
#include <IM.h>
#define CONV_RGB  1
#define CONV_UVW  2
#define CONV_YIQ  3
#define CONV_XYZ  4
#define CONV_UVW2 5
#define CONV_HIS  6
#define CONV_LAB  7

static float X_0 = 0.982;
static float Y_0 = 1.00;
static float Z_0 = 1.183;
static float u_0 = 0.201;
static float v_0 = 0.307;

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   COLOR_TYPE **Data1;
   FLOAT_TYPE ***Data2;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int Convert = CONV_RGB;
   int i = 0, j, k;
   float X, Y, Z, U, V, W, I, Q, R, G, B, U2, V2, W2;
   float H, S, L, a2, b2, u, v, temp, min;

   /* Interpret program options */
   fprintf(stdout, "COLOR_TO Program - KUIM Version 3.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'r':
	 Convert = CONV_RGB;
	 break;
      case 'y':
	 Convert = CONV_YIQ;
	 break;
      case 'u':
	 Convert = CONV_UVW;
	 break;
      case 'x':
	 Convert = CONV_XYZ;
	 break;
      case 'U':
	 Convert = CONV_UVW2;
	 break;
      case 'h':
	 Convert = CONV_HIS;
	 break;
      case 'l':
	 Convert = CONV_LAB;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: color_to [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-rgb]  Output is RGB format\n");
      fprintf(stderr, "       [-uvw]  Output is UVW format\n");
      fprintf(stderr, "       [-yiq]  Output is YIQ format\n");
      fprintf(stderr, "       [-xyz]  Output is XYZ format\n");
      fprintf(stderr, "       [-UVW]  Output is U*V*W* format\n");
      fprintf(stderr, "       [-his]  Output is HIS format\n");
      fprintf(stderr, "       [-lab]  Output is Lab format\n");
      exit(1);
   }
   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      fprintf(stderr, "Could not get input file name\n");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      fprintf(stderr, "Could not get output file name\n");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can only process 2D images");
   Data1 = (COLOR_TYPE **) im_alloc2D(Image1, COLOR);
   im_read(Image1, COLOR, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, FLOAT, Xdim, Ydim, 3);
   Data2 = (FLOAT_TYPE ***) im_alloc3D(Image2, FLOAT);

   /* Manipulate input image data */
   for (k = 0; k < Ydim; k++)
      for (j = 0; j < Xdim; j++)
      {
	 R = Data1[k][j].r;
	 G = Data1[k][j].g;
	 B = Data1[k][j].b;

	 if (Convert == CONV_RGB)
	 {
	    Data2[0][k][j] = R;
	    Data2[1][k][j] = G;
	    Data2[2][k][j] = B;
	 }
	 else if (Convert == CONV_UVW)
	 {
	    U = 0.405 * R + 0.116 * G + 0.133 * B;
	    V = 0.299 * R + 0.587 * G + 0.114 * B;
	    W = 0.145 * R + 0.827 * G + 0.627 * B;
	    Data2[0][k][j] = U;
	    Data2[1][k][j] = V;
	    Data2[2][k][j] = W;
	 }
	 else if (Convert == CONV_YIQ)
	 {
	    Y = 0.299 * R + 0.587 * G + 0.114 * B;
	    I = 0.596 * R - 0.274 * G - 0.322 * B;
	    Q = 0.211 * R - 0.523 * G + 0.312 * B;
	    Data2[0][k][j] = Y;
	    Data2[1][k][j] = I;
	    Data2[2][k][j] = Q;
	 }
	 else if (Convert == CONV_XYZ)
	 {
	    X = 0.607 * R + 0.174 * G + 0.201 * B;
	    Y = 0.299 * R + 0.587 * G + 0.114 * B;
	    Z = 0.066 * G + 1.117 * B;
	    Data2[0][k][j] = X;
	    Data2[1][k][j] = Y;
	    Data2[2][k][j] = Z;
	 }
	 else if (Convert == CONV_UVW2)
	 {
	    X = 0.607 * R + 0.174 * G + 0.201 * B;
	    Y = 0.299 * R + 0.587 * G + 0.114 * B;
	    Z = 0.066 * G + 1.117 * B;
	    temp = (X + 15.0 * Y + 3.0 * Z);
	    if (temp == 0)
	       u = 0;
	    else
	       u = 4.0 * X / temp;
	    if (temp == 0)
	       v = 0;
	    else
	       v = 6.0 * Y / temp;
	    W2 = 25.0 * (float) pow((double) (100.0 * Y), 1.0 / 3.0) - 17.0;
	    U2 = 13.0 * W2 * (u - u_0);
	    V2 = 13.0 * W2 * (v - v_0);
	    Data2[0][k][j] = U2;
	    Data2[1][k][j] = V2;
	    Data2[2][k][j] = W2;
	 }
	 else if (Convert == CONV_HIS)
	 {
#ifdef OLDCODE
	    X = 0.607 * R + 0.174 * G + 0.201 * B;
	    Y = 0.299 * R + 0.587 * G + 0.114 * B;
	    Z = 0.066 * G + 1.117 * B;
	    temp = (X + 15.0 * Y + 3.0 * Z);
	    if (temp == 0)
	       u = 0;
	    else
	       u = 4.0 * X / temp;
	    if (temp == 0)
	       v = 0;
	    else
	       v = 6.0 * Y / temp;
	    W2 = 25.0 * (float) pow((double) (100.0 * Y), 1.0 / 3.0) - 17.0;
	    U2 = 13.0 * W2 * (u - u_0);
	    V2 = 13.0 * W2 * (v - v_0);
	    S = (float) sqrt((double) (U2 * U2 + V2 * V2));
	    H = (float) atan2((double) (v - v_0), (double) (u - u_0));
	    if (H < 0.0)
	       H += (float) (2.0 * M_PI);
	    Data2[0][k][j] = H;
	    Data2[1][k][j] = W2;
	    Data2[2][k][j] = S;
#endif
	    /* Scale RGB to unit cube */
	    R /= 255;
	    G /= 255;
	    B /= 255;

	    /* Calculate hue */
	    temp = (R - G) * (R - G) + (R - B) * (G - B);
	    if (temp > 0)
	       Data2[0][k][j] = acos(0.5 * (2 * R - G - B) / sqrt(temp));
	    else
	       Data2[0][k][j] = 0;
	    if (G < B)
	       Data2[0][k][j] = 2 * M_PI - Data2[0][k][j];

	    /* Rotate hue by PI 
	       Data2[0][k][j] += M_PI;
	       if (Data2[0][k][j] > 2*M_PI)
	       Data2[0][k][j] -= M_PI;
	     */

	    /* Calculate intensity */
	    Data2[1][k][j] = (R + G + B) / 3.0;

	    /* Calculate saturation */
	    min = R;
	    if (G < min)
	       min = G;
	    if (B < min)
	       min = B;
	    temp = (R + G + B);
	    if (temp > 0)
	       Data2[2][k][j] = 3 * min / temp;
	    else
	       Data2[2][k][j] = 0;
	 }
	 else if (Convert == CONV_LAB)
	 {
	    X = 0.607 * R + 0.174 * G + 0.201 * B;
	    Y = 0.299 * R + 0.587 * G + 0.114 * B;
	    Z = 0.066 * G + 1.117 * B;
	    L = 25.0 * (float) pow((double) (100.0 * Y / Y_0), 1.0 / 3.0) - 16.0;
	    a2 = 500.0 * (float) pow((double) (X / X_0), 1.0 / 3.0) -
	       pow((double) (Y / Y_0), 1.0 / 3.0);
	    b2 = 200.0 * (float) pow((double) (Y / Y_0), 1.0 / 3.0) -
	       pow((double) (Z / Z_0), 1.0 / 3.0);
	    Data2[0][k][j] = L;
	    Data2[1][k][j] = a2;
	    Data2[2][k][j] = b2;
	 }
      }

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data2[0][0][0]));
   im_free2D((char **) Data1);
   im_free3D((char ***) Data2);
   return (0);
}
