/*---------------------------------------------------------------------------*/
/* Program:  convert_raw.c                                                   */
/*                                                                           */
/* Purpose:  This program converts raw data files into KUIM image format.    */
/*           A variable size header can be skipped and a variety of          */
/*           pixel types are supported.  No byte swapping is performed.      */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     May 6, 1994       - Original program.                           */
/*           August 1, 1997    - Added support for ascii.                    */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define ASCII 42

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   BYTE_TYPE *ByteData;
   SHORT_TYPE *ShortData;
   INT_TYPE *IntData;
   FLOAT_TYPE *FloatData;
   COLOR_TYPE *ColorData;
   FILE *Fd;

   /* Program variables */
   int Xdim = 256, Ydim = 256, Zdim = 1;
   int PixType = BYTE;
   int Debug = FALSE;
   int Count;
   int Skip = 0;
   int i = 0;

   /* Interpret program options */
   printf("CONVERT_RAW Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 's':
	 if (sscanf(argv[++i], "%d", &Skip) == 0)
	    Error("Could not get Skip argument");
	 break;
      case 'x':
	 if (sscanf(argv[++i], "%d", &Xdim) == 0)
	    Error("Could not get Xdim argument");
	 break;
      case 'y':
	 if (sscanf(argv[++i], "%d", &Ydim) == 0)
	    Error("Could not get Ydim argument");
	 break;
      case 'z':
	 if (sscanf(argv[++i], "%d", &Zdim) == 0)
	    Error("Could not get Zdim argument");
	 break;
      case 'B':
	 PixType = BYTE;
	 break;
      case 'S':
	 PixType = SHORT;
	 break;
      case 'I':
	 PixType = INT;
	 break;
      case 'F':
	 PixType = FLOAT;
	 break;
      case 'C':
	 PixType = COLOR;
	 break;
      case 'A':
	 PixType = ASCII;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: convert_raw [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-s #]  Number of header bytes to skip (0)\n");
      fprintf(stderr, "       [-x #]  Size of X dimension (256)\n");
      fprintf(stderr, "       [-y #]  Size of Y dimension (256)\n");
      fprintf(stderr, "       [-z #]  Size of Z dimension (1)\n");
      fprintf(stderr, "       [-B]    Input file is BYTE type\n");
      fprintf(stderr, "       [-S]    Input file is SHORT type\n");
      fprintf(stderr, "       [-I]    Input file is INT type\n");
      fprintf(stderr, "       [-F]    Input file is FLOAT type\n");
      fprintf(stderr, "       [-C]    Input file is COLOR type\n");
      fprintf(stderr, "       [-A]    Input file is ASCII type\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Open input image and skip header */
   Count = Xdim * Ydim * Zdim;
   if ((Fd = fopen(Name1, "r")) == NULL)
      Error("Can not open image file");
   if ((Skip > 0) && (PixType != ASCII))
   {
      ByteData = (BYTE_TYPE *) malloc((unsigned) Skip);
      fread(ByteData, 1, Skip, Fd);
      if (Debug == TRUE)
	 printf("skipping %d bytes\n", Skip);
   }

   /* Handle ASCII image */
   if (PixType == ASCII)
   {
      /* Read image dimensions */
      if (fscanf(Fd, "%d %d %d", &Xdim, &Ydim, &Zdim) != 3)
	 Error("Could not read image dimensions");
      if (Debug == TRUE)
	 printf("image dimensions = %d %d %d\n", Xdim, Ydim, Zdim);

      /* Create output image */
      Count = Xdim * Ydim * Zdim;
      Image1 = im_create(Name2, FLOAT, Xdim, Ydim, Zdim);
      FloatData = (FLOAT_TYPE *) im_alloc1D(Image1, FLOAT);

      /* Read image data */
      for (i = 0; i < Count; i++)
	 if (fscanf(Fd, "%f", &(FloatData[i])) != 1)
	    Error("Could not read ascii value");

      /* Write output image */
      im_write(Image1, FLOAT, (char *) FloatData);
      im_free1D((char *) FloatData);
   }

   /* Handle BYTE image */
   else if (PixType == BYTE)
   {
      /* Use SHORT instead of PixType to turn OFF data scaling */
      Image1 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
      ByteData = (BYTE_TYPE *) im_alloc1D(Image1, BYTE);
      fread(ByteData, sizeof(BYTE_TYPE), Count, Fd);
      im_write(Image1, BYTE, (char *) ByteData);
      im_free1D((char *) ByteData);
   }

   /* Handle SHORT image */
   else if (PixType == SHORT)
   {
      Image1 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
      ShortData = (SHORT_TYPE *) im_alloc1D(Image1, SHORT);
      fread(ShortData, sizeof(SHORT_TYPE), Count, Fd);
      im_write(Image1, SHORT, (char *) ShortData);
      im_free1D((char *) ShortData);
   }

   /* Handle INT image */
   else if (PixType == INT)
   {
      Image1 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
      IntData = (INT_TYPE *) im_alloc1D(Image1, INT);
      fread(IntData, sizeof(INT_TYPE), Count, Fd);
      im_write(Image1, INT, (char *) IntData);
      im_free1D((char *) IntData);
   }

   /* Handle FLOAT image */
   else if (PixType == FLOAT)
   {
      Image1 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
      FloatData = (FLOAT_TYPE *) im_alloc1D(Image1, FLOAT);
      fread(FloatData, sizeof(FLOAT_TYPE), Count, Fd);
      im_write(Image1, FLOAT, (char *) FloatData);
      im_free1D((char *) FloatData);
   }

   /* Handle COLOR image */
   else if (PixType == COLOR)
   {
      Image1 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
      ColorData = (COLOR_TYPE *) im_alloc1D(Image1, COLOR);
      fread(&(ColorData[0].r), 3, Xdim * Ydim, Fd);
      im_write(Image1, COLOR, (char *) ColorData);
      im_free1D((char *) ColorData);
   }

   /* Close input image */
   fclose(Fd);
   return (0);
}
