/*---------------------------------------------------------------------------*/
/* Program:  deform2d.c                                                      */
/*                                                                           */
/* Purpose:  This program reads in a 2D contour and deforms it to match      */
/*           high gradient features in an imput image.                       */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     September 14, 1994                                              */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Global variables */
FLOAT_TYPE **DataIn;
FLOAT_TYPE **DataOut;
FLOAT_TYPE **Attract;
int Xdim, Ydim, Zdim, Udim, Vdim;

/* Deformation parameters */
float alpha = 0.2;
float beta = 0.1;
float delta = -10;
float step = 0.1;
int num_iter = 100;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine deforms a 2D boundary.                             */
/*---------------------------------------------------------------------------*/
void Deform2D(int num_iter)
{
   int iteration, index;
   int X, Y, L1, L2, R1, R2;
   float Gx, Xuu, Xuuuu, Xerr;
   float Gy, Yuu, Yuuuu, Yerr;
   float *Xold, *Yold, *Xnew, *Ynew;

   /* Set up pointers */
   Xold = DataIn[0];
   Yold = DataIn[1];
   Xnew = DataOut[0];
   Ynew = DataOut[1];

   /* Deform object boundary */
   for (iteration = 0; iteration < num_iter; iteration++)
   {
      /* Deform boundary */
      for (index = 0; index < Udim; index++)
      {
	 /* Set up indices */
	 L1 = (index + Udim - 1) % Udim;
	 L2 = (index + Udim - 2) % Udim;
	 R1 = (index + Udim + 1) % Udim;
	 R2 = (index + Udim + 2) % Udim;

	 /* Calculate gradient of attract function */
	 X = (int) Xold[index];
	 Y = (int) Yold[index];
	 Gx = Attract[Y][X] - Attract[Y][X - 1];
	 Gy = Attract[Y][X] - Attract[Y - 1][X];

	 /* Update x,y coordinates */
	 Xuu = Xold[L1] - 2 * Xold[index] + Xold[R1];
	 Yuu = Yold[L1] - 2 * Yold[index] + Yold[R1];
	 Xuuuu = Xold[L2] - 4 * Xold[L1] + 6 * Xold[index] - 4 * Xold[R1] + Xold[R2];
	 Yuuuu = Yold[L2] - 4 * Yold[L1] + 6 * Yold[index] - 4 * Yold[R1] + Yold[R2];
	 Xerr = delta * Gx - alpha * Xuu + beta * Xuuuu;
	 Yerr = delta * Gy - alpha * Yuu + beta * Yuuuu;
	 Xnew[index] = Xold[index] - step * Xerr;
	 Ynew[index] = Yold[index] - step * Yerr;

	 /* Check for range errors */
	 if (Xnew[index] < 1)
	    Xnew[index] = 1.0;
	 if (Ynew[index] < 1)
	    Ynew[index] = 1.0;
	 if (Xnew[index] > Xdim - 2)
	    Xnew[index] = (float) (Xdim - 2);
	 if (Ynew[index] > Ydim - 2)
	    Ynew[index] = (float) (Ydim - 2);
      }

      /* Copy boundary */
      for (index = 0; index < Udim; index++)
      {
	 Xold[index] = Xnew[index];
	 Yold[index] = Ynew[index];
      }
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   int PixType, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int i = 0;

   /* Interpret program options */
   printf("DEFORM2D Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'n':
	 if (sscanf(argv[++i], "%d", &num_iter) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr, "Usage: deform2d [options] attrfile infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-n #]  Number of iterations (100)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get attract file name");
   if (sscanf(argv[i++], "%s", Name3) == 0)
      Error("Could not get contour file name");

   /* Read attract image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can only process 2D images");
   Attract = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Attract[0][0]));

   /* Read input contour */
   Image2 = im_open(Name2, &PixType, &Udim, &Vdim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can only process 2D images");
   DataIn = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   im_read(Image2, FLOAT, (char *) &(DataIn[0][0]));

   /* Create output contour */
   Image3 = im_create(Name3, PixType, Udim, Vdim, Zdim);
   DataOut = (FLOAT_TYPE **) im_alloc2D(Image3, FLOAT);

   /* Perform deformation */
   Deform2D(num_iter);

   /* Write output contour */
   im_write(Image3, FLOAT, (char *) &(DataIn[0][0]));
   im_free2D((char **) Attract);
   im_free2D((char **) DataIn);
   im_free2D((char **) DataOut);
   return (0);
}
