/*---------------------------------------------------------------------------*/
/* Program:  deform3d.c                                                      */
/*                                                                           */
/* Purpose:  This program reads in a 3D boundary and deforms it to match     */
/*           high gradient features in an imput image.                       */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     September 15, 1994                                              */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Global variables */
int Debug = FALSE;
FLOAT_TYPE ***DataIn;
FLOAT_TYPE ***DataOut;
FLOAT_TYPE ***Attract;
int Xdim, Ydim, Zdim, Udim, Vdim;

/* Deformation parameters */
float alpha = 0.2;
float beta = 0.1;
float delta = -10;
float step = 0.1;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine deforms a 3D boundary.                             */
/*---------------------------------------------------------------------------*/
void Deform3D(int num_iter)
{
   int iteration, u, v, X, Y, Z, L1, L2, R1, R2, U1, U2, D1, D2;
   float Gx, Xuu, Xvv, Xuuuu, Xuuvv, Xvvvv, Xerr;
   float Gy, Yuu, Yvv, Yuuuu, Yuuvv, Yvvvv, Yerr;
   float **Xold, **Yold, **Zold, **Xnew, **Ynew, **Znew;

   /* Set up pointers */
   Xold = DataIn[0];
   Yold = DataIn[1];
   Zold = DataIn[2];
   Xnew = DataOut[0];
   Ynew = DataOut[1];
   Znew = DataOut[2];

   /* Deform object boundary */
   for (iteration = 0; iteration < num_iter; iteration++)
   {
      /* Deform boundary */
      for (v = 0; v < Vdim; v++)
	 for (u = 0; u < Udim; u++)
	 {
	    /* Set up indices */
	    L1 = (u + Udim - 1) % Udim;
	    L2 = (u + Udim - 2) % Udim;
	    R1 = (u + Udim + 1) % Udim;
	    R2 = (u + Udim + 2) % Udim;
	    U1 = (v < 1) ? 0 : v - 1;
	    U2 = (v < 2) ? 0 : v - 2;
	    D1 = (v > Vdim - 2) ? Vdim - 1 : v + 1;
	    D2 = (v > Vdim - 3) ? Vdim - 1 : v + 2;

	    /* Calculate gradient of attract function */
	    X = (int) Xold[v][u];
	    Y = (int) Yold[v][u];
	    Z = (int) Zold[v][u];
	    Gx = Gy = 0.0;
	    if (X > 0 && Y > 0 && X < Xdim && Y < Ydim)
	    {
	       Gx = Attract[Z][Y][X] - Attract[Z][Y][X - 1];
	       Gy = Attract[Z][Y][X] - Attract[Z][Y - 1][X];
	    }

	    /* Calculate surface derivatives */
	    Xuu = Xold[v][L1] - 2 * Xold[v][u] + Xold[v][R1];
	    Xvv = Xold[U1][u] - 2 * Xold[v][u] + Xold[D1][u];
	    Xuuuu = Xold[v][L2] - 4 * Xold[v][L1] + 6 * Xold[v][u] - 4 * Xold[v][R1] + Xold[v][R2];
	    Xvvvv = Xold[U2][u] - 4 * Xold[U1][u] + 6 * Xold[v][u] - 4 * Xold[D1][u] + Xold[D2][u];
	    Xuuvv = Xold[U1][L1] - 2 * Xold[U1][u] + Xold[U1][R1]
	       - 2 * Xold[v][L1] + 4 * Xold[v][u] - 2 * Xold[v][R1]
	       + Xold[D1][L1] - 2 * Xold[D1][u] + Xold[D1][R1];
	    Yuu = Yold[v][L1] - 2 * Yold[v][u] + Yold[v][R1];
	    Yvv = Yold[U1][u] - 2 * Yold[v][u] + Yold[D1][u];
	    Yuuuu = Yold[v][L2] - 4 * Yold[v][L1] + 6 * Yold[v][u] - 4 * Yold[v][R1] + Yold[v][R2];
	    Yvvvv = Yold[U2][u] - 4 * Yold[U1][u] + 6 * Yold[v][u] - 4 * Yold[D1][u] + Yold[D2][u];
	    Yuuvv = Yold[U1][L1] - 2 * Yold[U1][u] + Yold[U1][R1]
	       - 2 * Yold[v][L1] + 4 * Yold[v][u] - 2 * Yold[v][R1]
	       + Yold[D1][L1] - 2 * Yold[D1][u] + Yold[D1][R1];

	    /* Update x,y coordinates */
	    Xerr = delta * Gx - alpha * (Xuu + Xvv) + beta * (Xuuuu + 2 * Xuuvv + Xvvvv);
	    Yerr = delta * Gy - alpha * (Yuu + Yvv) + beta * (Yuuuu + 2 * Yuuvv + Yvvvv);
	    Xnew[v][u] = Xold[v][u] - step * Xerr;
	    Ynew[v][u] = Yold[v][u] - step * Yerr;

	    /* Check for range errors */
	    if (Xnew[v][u] < 1)
	       Xnew[v][u] = 1.0;
	    if (Ynew[v][u] < 1)
	       Ynew[v][u] = 1.0;
	    if (Xnew[v][u] > Xdim - 2)
	       Xnew[v][u] = (float) (Xdim - 2);
	    if (Ynew[v][u] > Ydim - 2)
	       Ynew[v][u] = (float) (Ydim - 2);
	 }

      /* Copy boundary */
      for (v = 0; v < Vdim; v++)
	 for (u = 0; u < Udim; u++)
	 {
	    Xold[v][u] = Xnew[v][u];
	    Yold[v][u] = Ynew[v][u];
	 }
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   int PixType, DimCnt;

   /* Program variables */
   int num_iter = 100;
   int i = 0;

   /* Interpret program options */
   printf("DEFORM3D Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'n':
	 if (sscanf(argv[++i], "%d", &num_iter) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr, "Usage: deform3d [options] attrfile infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-n #]  Number of iterations (100)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get attract file name");
   if (sscanf(argv[i++], "%s", Name3) == 0)
      Error("Could not get contour file name");

   /* Read attract image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 3)
      Error("Can only process 3D images");
   Attract = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Attract[0][0][0]));

   /* Read input contour */
   Image2 = im_open(Name2, &PixType, &Udim, &Vdim, &Zdim, &DimCnt);
   if (DimCnt != 3)
      Error("Can only process 3D images");
   DataIn = (FLOAT_TYPE ***) im_alloc3D(Image2, FLOAT);
   im_read(Image2, FLOAT, (char *) &(DataIn[0][0][0]));

   /* Create output contour */
   Image3 = im_create(Name3, PixType, Udim, Vdim, Zdim);
   DataOut = (FLOAT_TYPE ***) im_alloc3D(Image3, FLOAT);

   /* Perform deformation */
   Deform3D(num_iter);

   /* Write output contour */
   im_write(Image3, FLOAT, (char *) &(DataIn[0][0][0]));
   im_free3D((char ***) Attract);
   im_free3D((char ***) DataIn);
   im_free3D((char ***) DataOut);
   return (0);
}
