/*---------------------------------------------------------------------------*/
/* Program:  filter_project.c                                                */
/*                                                                           */
/* Purpose:  This program implements the filtering stage of the filtered     */
/*           back projection image reconstruction technique.  Each row of    */
/*           the input image is converted to the frequency domain via FFT,   */
/*           filtered by |w|, and converted back to the spatial domain       */
/*           via an inverse FFT. An exponential falloff has been added       */
/*           to the filter to lessen the effect of aliasing artifacts.       */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     May 11, 1994                                                    */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Program variables */
int Debug = FALSE;
int Width = 200;
float StdDev = 10.0;
int Xdim, Ydim, Zdim;
COMPLEX_TYPE **Data1;
double exp();

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine filters each row in the imput image.               */
/*---------------------------------------------------------------------------*/
void Filter()
{
   int i, a, b;
   float *filter, exponent;

   /* Initialize the filter */
   filter = (float *) malloc(Xdim * sizeof(float));
   if (Width > Xdim / 2)
      Width = Xdim / 2;
   filter[0] = 0;
   for (i = 1; i < Width; i++)
   {
      filter[Xdim - i] = filter[i] = (float) i;
      if (Debug == TRUE)
	 printf("filter[%d] = %f\n", i, filter[i]);
   }
   for (i = Width; i <= Xdim / 2; i++)
   {
      exponent = -(Width - i) * (Width - i) / (StdDev * StdDev);
      filter[Xdim - i] = filter[i] = Width * (float) exp((double) exponent);
      if (Debug == TRUE)
	 printf("filter[%d] = %f\n", i, filter[i]);
   }

   /* Filter each projection */
   for (a = 0; a < Ydim; a++)
   {
      if (Debug == TRUE)
	 printf("row %d\n", a);
      fft_1d(&(Data1[a][0]), Xdim, -1);
      for (b = 0; b < Xdim; b++)
      {
	 Data1[a][b].re *= filter[b];
	 Data1[a][b].im *= filter[b];
      }
      fft_1d(&(Data1[a][0]), Xdim, 1);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   int PixType, DimCnt;

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("FILTER_PROJECT Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'w':
	 if (sscanf(argv[++i], "%d", &Width) == 0)
	    Error("Could not get Width argument");
	 break;
      case 's':
	 if (sscanf(argv[++i], "%f", &StdDev) == 0)
	    Error("Could not get StdDev argument");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: filter_project [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-w #]  Width of ideal filter (200)\n");
      fprintf(stderr, "       [-s #]  StdDev of falloff function (10)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (COMPLEX_TYPE **) im_alloc2D(Image1, COMPLEX);
   im_read(Image1, COMPLEX, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Filter();

   /* Write information to output image */
   im_write(Image2, COMPLEX, (char *) &(Data1[0][0]));
   im_free2D((char **) Data1);
   return (0);
}
