/*---------------------------------------------------------------------------*/
/* Program:  hierarchy.c                                                     */
/*                                                                           */
/* Purpose:  This program displays images and associated region hierarchies. */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     September 1, 1996 - Rewrote old /usr/image program.             */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1996                    */
/*---------------------------------------------------------------------------*/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <IM.h>

/* Global variables */
#define SIZE 50000
int Parent[SIZE];
int Level[SIZE];
char Show[SIZE];
int RegionCnt = 0;
int Debug = FALSE;

/*---------------------------------------------------------------------------*/
/* Purpose: This routine displays a 2D image.                                */
/*---------------------------------------------------------------------------*/
Display2D(
	    char *Name,
	    BYTE_TYPE * Save,
	    BYTE_TYPE * Data,
	    SHORT_TYPE * Region,
	    BYTE_TYPE * Mask,
	    int Xdim,
	    int Ydim)
{
   Display *display;
   char *display_name = NULL;
   int screen_num;
   int depth;
   unsigned long foreground;
   unsigned long background;
   Window window;
   GC gc;
   Visual *visual;
   Colormap old_colormap;
   Colormap new_colormap;
   int i, x, y, region, child, level;
   XColor color;
   double Gamma = 0.01;
   XImage *ximage;
   int done = FALSE;
   XEvent event;
   char text[50];
   KeySym key;
   IM_TYPE *Image;

   /* Set up window for display */
   if ((display = XOpenDisplay(display_name)) == NULL)
      Error("Can not connect to X server");
   screen_num = DefaultScreen(display);
   depth = DefaultDepth(display, screen_num);
   if (depth != 8)
      Error("Can only display on 8 bit displays");
   background = BlackPixel(display, screen_num);
   foreground = WhitePixel(display, screen_num);
   window = XCreateSimpleWindow(display, DefaultRootWindow(display),
			   200, 200, Xdim, Ydim, 5, foreground, background);
   XStoreName(display, window, Name);
   XSelectInput(display, window, ButtonPressMask | KeyPressMask | ExposureMask);
   gc = XCreateGC(display, window, 0, 0);
   XSetBackground(display, gc, background);
   XSetForeground(display, gc, foreground);
   visual = XDefaultVisualOfScreen(DefaultScreenOfDisplay(display));
   XMapRaised(display, window);

   /* Define colormap */
   old_colormap = DefaultColormap(display, screen_num);
   new_colormap = XCreateColormap(display, window, visual, AllocAll);
   for (i = 0; i <= 255; i++)
   {
      color.pixel = i;
      XQueryColor(display, old_colormap, &color);
      color.flags = DoRed | DoGreen | DoBlue;
      XStoreColor(display, new_colormap, &color);
   }
   for (i = 50; i <= 250; i++)
   {
      color.pixel = i;
      color.red = color.blue = color.green = (int) (65535
				    * log((double) (1.0 + Gamma * (i - 50)))
				       / log((double) (1.0 + Gamma * 200)));
      XStoreColor(display, new_colormap, &color);
   }

   /* Add GREEN entry for marking points */
   color.pixel = 251;
   color.green = 255 * 255;
   color.red = 255 * 255;
   color.blue = 255 * 255;
   XStoreColor(display, new_colormap, &color);

   /* Copy image to display window */
   XSetWindowColormap(display, window, new_colormap);
   ximage = XCreateImage(display, visual, 8, ZPixmap, 0, Data, Xdim, Ydim, 8, 0);

   /* Loop handling events */
   while (done == FALSE)
   {
      XNextEvent(display, &event);
      switch (event.type)
      {
      case Expose:
	 if (event.xexpose.count == 0)
	    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	 break;
      case MappingNotify:
	 XRefreshKeyboardMapping((XMappingEvent *) & event);
	 break;

      case ButtonPress:
	 switch (event.xbutton.button)
	 {
	 case 1:		/* Select starting region */
	    level = 0;
	    region = Region[event.xbutton.y * Xdim + event.xbutton.x];
	    break;

	 case 2:		/* Move up one blur level */
	    level++;
	    if ((level >= Level[region]) && (Parent[region] > 0))
	    {
	       level = Level[region];
	       region = Parent[region];
	    }
	    for (i = 0; i < RegionCnt; i++)
	    {
	       child = i;
	       while ((Parent[child] > 0) &&
		      (Parent[child] != region))
		  child = Parent[child];
	       if ((Parent[child] == region) &&
		   (Level[child] <= level))
	       {
		  if (Debug == TRUE)
		     printf("child %d level %d\n",
			    i, Level[child]);
		  Show[i] = TRUE;
	       }
	    }
	    break;

	 case 3:		/* Move up to parent region */
	    if (Parent[region] > 0)
	       region = Parent[region];
	    level = Level[region];
	    for (i = 0; i < RegionCnt; i++)
	    {
	       child = i;
	       while ((Parent[child] > 0) &&
		      (Parent[child] != region))
		  child = Parent[child];
	       if ((Parent[child] == region) &&
		   (Level[child] <= level))
	       {
		  if (Debug == TRUE)
		     printf("child %d level %d\n",
			    i, Level[child]);
		  Show[i] = TRUE;
	       }
	    }
	    break;
	 }

	 /* Mark all pixels to be displayed */
	 Show[region] = TRUE;
	 for (i = 0; i < Xdim * Ydim; i++)
	    if (Show[Region[i]] == TRUE)
	       Mask[i] = 1;

	 /* Draw boundaries of mask */
	 for (y = 1; y < Ydim; y++)
	    for (x = 1; x < Xdim; x++)
	    {
	       i = y * Xdim + x;
	       if ((Mask[i] == 1) && (Mask[i - 1] != 1))
		  Data[i] = 251;
	       else if ((Mask[i] != 1) && (Mask[i - 1] == 1))
		  Data[i] = 251;
	       else if ((Mask[i] == 1) && (Mask[i - Xdim] != 1))
		  Data[i] = 251;
	       else if ((Mask[i] != 1) && (Mask[i - Xdim] == 1))
		  Data[i] = 251;
	       else
		  Data[i] = Save[i];
	    }

	 /* Put image in window */
	 printf("(%d, %d) region=%d level=%d\n", event.xbutton.x,
		event.xbutton.y, region, level);
	 XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	 break;

      case KeyPress:
	 /* Exit if 'q' was pressed */
	 i = XLookupString((XKeyEvent *) & event, text, 50, &key, 0);
	 if (i == 1 && text[0] == 'q')
	    done = TRUE;

	 /* Save mask */
	 Image = im_create("mask", BYTE, 256, 256, 1);
	 im_write(Image, BYTE, (char *) Mask);
	 break;
      }
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   BYTE_TYPE *Data0;
   BYTE_TYPE *Data1;
   SHORT_TYPE *Data2;
   BYTE_TYPE *Data3;
   int Xdim, Ydim, Zdim, DimCnt, PixType;

   /* Program variables */
   int Count, Min, Max;
   int i = 0, x, y, z;
   int child, parent, level;
   FILE *Fd;

   /* Interpret program options */
   fprintf(stderr, "HIERARCHY Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr, "Usage: hierarchy [options] infile regionfile datafile\n");
      fprintf(stderr, "       [-d] Print debugging information\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get region file name");
   if (sscanf(argv[i++], "%s", Name3) == 0)
      Error("Could not get data file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   Data0 = (BYTE_TYPE *) im_alloc1D(Image1, BYTE);
   Data1 = (BYTE_TYPE *) im_alloc1D(Image1, BYTE);

   /* Read region image */
   Image2 = im_open(Name2, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   Data2 = (SHORT_TYPE *) im_alloc1D(Image2, SHORT);
   Data3 = (BYTE_TYPE *) im_alloc1D(Image2, BYTE);
   im_read(Image2, SHORT, (char *) Data2);

   /* Read hierarchy data file */
   if ((Fd = fopen(Name3, "r")) == NULL)
      Error("Can not open hierarchy Data file");
   while (fscanf(Fd, "%d %d %d", &child, &parent, &level) != EOF)
   {
      Parent[child] = parent;
      Level[child] = level;
      RegionCnt++;
   }
   fclose(Fd);
   printf("Region Count = %d\n", RegionCnt);

   /* Calculate range of pixel values */
   im_read(Image1, BYTE, (char *) Data1);
   Count = Xdim * Ydim * Zdim;
   Min = Max = Data1[0];
   for (i = 0; i < Count; i++)
      if (Min > Data1[i])
	 Min = Data1[i];
      else if (Max < Data1[i])
	 Max = Data1[i];

   /* Rescale to 50..200 range */
   for (i = 0; i < Count; i++)
      if (Max > Min)
	 Data0[i] = Data1[i] = 50 + ((Data1[i] - Min) * 200) / (Max - Min);
      else
	 Data0[i] = Data1[i] = 50;

   /* Display image */
   if (DimCnt == 2)
      Display2D(Name1, Data0, Data1, Data2, Data3, Xdim, Ydim);
   im_free1D((char *) Data0);
   im_free1D((char *) Data1);
   im_free1D((char *) Data2);
   exit(0);
}
