/*---------------------------------------------------------------------------*/
/* Program:  huff_decode.c                                                   */
/*                                                                           */
/* Purpose:  This program performs Huffman decoding of a compressed          */
/*           binary file to obtain a KUIM image.  The original image         */
/*           dimensions and histogram are read from the file and             */
/*           used to regenerate the original Huffman code tree.              */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     October 28, 1995 - Original program.                            */
/*           November 1, 1995 - Extended to any grey level range.            */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define MAX_SIZE 10000

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   SHORT_TYPE **Data1;
   int Xdim, Ydim;

   /* Program variables */
   int Debug = FALSE;
   int i = 0, x, y;
   char Code;
   short int Prob[MAX_SIZE];
   int Parent[MAX_SIZE];
   int Zero[MAX_SIZE];
   int One[MAX_SIZE];
   int Loop, Min, Max, Range, Pos;
   FILE *fd;
   int Buffer;
   int BufLen;

   /* Interpret program options */
   fprintf(stderr, "HUFF_DECODE Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: huff_decode [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read Huffman file header */
/*--------------------------*/
   for (i = 0; i < MAX_SIZE; i++)
   {
      Prob[i] = 0;
      Parent[i] = -1;
      Zero[i] = -1;
      One[i] = -1;
   }
   if ((fd = fopen(Name1, "r")) == NULL)
      Error("Could not open input file");
   fread(&Xdim, 1, sizeof(Xdim), fd);
   fread(&Ydim, 1, sizeof(Ydim), fd);
   fread(&Min, 1, sizeof(Min), fd);
   fread(&Max, 1, sizeof(Max), fd);
   Range = Max - Min + 1;
   fread(Prob, Range, sizeof(Prob[0]), fd);

   /* Create output image */
/*---------------------*/
   Image1 = im_create(Name2, SHORT, Xdim, Ydim, 1);
   Data1 = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);

   /* Build the Huffman code tree */
/*-----------------------------*/
   for (Loop = 0; Loop < (Range - 1); Loop++)
   {
      /* Process lowest probability symbol */
      Pos = -1;
      Min = Xdim * Ydim;
      for (i = 0; i < Range + Loop; i++)
	 if ((Prob[i] < Min) && (Parent[i] == -1))
	 {
	    Min = Prob[i];
	    Pos = i;
	 }
      Prob[Range + Loop] += Min;
      Zero[Range + Loop] = Pos;
      Parent[Pos] = Range + Loop;

      /* Process second lowest probability symbol */
      Pos = -1;
      Min = Xdim * Ydim;
      for (i = 0; i < Range + Loop; i++)
	 if ((Prob[i] < Min) && (Parent[i] == -1))
	 {
	    Min = Prob[i];
	    Pos = i;
	 }
      Prob[Range + Loop] += Min;
      One[Range + Loop] = Pos;
      Parent[Pos] = Range + Loop;
   }

   /* Decode image data using Huffman codes */
/*---------------------------------------*/
   fread(&Buffer, 1, sizeof(Buffer), fd);
   BufLen = 32;
   Min = Max - Range + 1;
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 /* Start at root of tree and work towards leaf */
	 Pos = 2 * Range - 2;
	 while (Zero[Pos] >= 0)
	 {
	    /* Read bit from compressed file */
	    if (Buffer < 0)
	       Code = 1;
	    else
	       Code = 0;
	    Buffer = Buffer << 1;
	    BufLen--;
	    if (BufLen == 0)
	    {
	       fread(&Buffer, 1, sizeof(Buffer), fd);
	       BufLen = 32;
	    }

	    /* Follow path through tree */
	    if (Code == 0)
	       Pos = Zero[Pos];
	    else if (Code == 1)
	       Pos = One[Pos];
	 }
	 Data1[y][x] = Pos + Min;
      }

   /* Write information to output image */
/*-----------------------------------*/
   im_write(Image1, SHORT, (char *) &(Data1[0][0]));
   im_free2D((char **) Data1);
   fclose(fd);
   return (0);
}
