/*---------------------------------------------------------------------------*/
/* Program:  inv_radon.c                                                     */
/*                                                                           */
/* Purpose:  This program calculates the inverse Radon transform of a        */
/*           projection image to obtain the original image.  This is         */
/*           is accomplished by:                                             */
/*           1) take 1D inverse FFT of each row of the projection image      */
/*           2) nearest neighbor interpolate from F(r,t) to F(u,v)           */
/*           3) take 2D inverse FFT of F(u,v)                                */
/*           4) translate f(x,y) to center and trim output image             */
/*           The output results will be similar to using FILTER_PROJECT      */
/*           and BACK_PROJECT except there may be high frequency artifacts.  */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     November 11, 1994                                               */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   COMPLEX_TYPE **Data1;
   COMPLEX_TYPE **Data2;
   FLOAT_TYPE **Data3;
   int PixType, Rdim, Tdim, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int i = 0, t, r, x, y, x1, y1;
   int halfx, halfy;

   /* Interpret program options */
   printf("INV_RADON Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: inv_radon [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Rdim, &Tdim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (COMPLEX_TYPE **) im_alloc2D(Image1, COMPLEX);
   im_read(Image1, COMPLEX, (char *) &(Data1[0][0]));

   /* Create temporary image */
   Xdim = Ydim = Rdim;
   Image2 = im_create("/dev/null", COMPLEX, Xdim, Ydim, Zdim);
   Data2 = (COMPLEX_TYPE **) im_alloc2D(Image2, COMPLEX);

   /* Create output image */
   Image3 = im_create(Name2, FLOAT, Xdim / 2, Ydim / 2, Zdim);
   Data3 = (FLOAT_TYPE **) im_alloc2D(Image3, FLOAT);

   /* Calculate 1D fourier transforms */
   for (t = 0; t < Tdim; t++)
      fft_1d(&(Data1[t][0]), Rdim, -1);

   /* Generate output image */
   halfx = Xdim / 2;
   halfy = Ydim / 2;
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 /* Wrap around x,y coordinates */
	 if (y < halfy)
	    y1 = y;
	 else
	    y1 = y - Ydim;
	 if (x < halfx)
	    x1 = x;
	 else
	    x1 = x - Xdim;

	 /* Calculate rounded r,t coordinates */
	 if (x1 == 0 && y1 == 0)
	    r = t = 0;
	 else
	 {
	    r = (int) (sqrt((double) x1 * x1 + y1 * y1) + 0.5);
	    t = (int) (atan2((double) y1, (double) x1) * Tdim / M_PI + 0.5);
	    if (t < 0)
	    {
	       t = Tdim + t;
	       r = Rdim - r;
	    }
	 }

	 /* Copy data to F(u,v) image */
	 if ((t >= 0) && (t < Tdim) && (r >= 0) && (r < Rdim))
	 {
	    Data2[y][x].im = Data1[t][r].im;
	    Data2[y][x].re = Data1[t][r].re;
	 }
	 else
	 {
	    Data2[y][x].im = 0;
	    Data2[y][x].re = 0;
	 }
      }

   /* Take 2D inverse FFT */
   fft_2d((char *) &(Data2[0][0]), Xdim, Ydim, 1);

   /* Shift origin of output image */
   for (y = 0; y < halfy; y++)
      for (x = 0; x < halfx; x++)
	 Data3[y][x] = Data2[(y + halfy + halfy / 2) % Ydim][(x + halfx + halfx / 2) % Xdim].re;

   /* Write information to output image */
   im_write(Image3, FLOAT, (char *) &(Data3[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   im_free2D((char **) Data3);
   return (0);
}
