/*---------------------------------------------------------------------------*/
/*  These functions are from the Numerical Recipes in C book.                */
/*  Includes routines for creation of vectors and matrices.                  */
/*  Matrices are represented by an array of pointers to the                  */
/*  start of each row, e.g., **a = *a[] . This allows a matrix               */
/*  to be passed to a subroutine as sub(a,m,n) and then                      */
/*  referenced in the routine as a[i][j] (a declared float **a;).            */
/*  Hence arbitrary size arrays are handled. The a[i][j] reference           */
/*  also only takes two adds and two references and no multiplies.           */
/*  Written by Larry Lifshitz 1990, edited by John Gauch 1995.               */
/*---------------------------------------------------------------------------*/
#define TINY 1.0e-20;

/* allocates a float vector in the range [nl...nh] */
float *vector(int nl, int nh)
{
   float *v;
   v = (float *) malloc((unsigned) (nh - nl + 1) * sizeof(float));
   if (!v)
      Error("allocation failure in vector()");
   return (v - nl);
}

/* allocates a integer vector in the range [nl...nh] */
int *ivector(int nl, int nh)
{
   int *v;
   v = (int *) malloc((unsigned) (nh - nl + 1) * sizeof(int));
   if (!v)
      Error("allocation failure in ivector()");
   return (v - nl);
}

/* allocates a double vector in the range [nl...nh] */
double *dvector(int nl, int nh)
{
   double *v;
   v = (double *) malloc((unsigned) (nh - nl + 1) * sizeof(double));
   if (!v)
      Error("allocation failure in dvector()");
   return (v - nl);
}

/* allocates a float matrix with range [nrl...nrh][ncl...nch]  */
float **matrix(int nrl, int nrh, int ncl, int nch)
{
   int i;
   float **m;
   m = (float **) malloc((unsigned) (nrh - nrl + 1) * sizeof(float *));
   if (!m)
      Error("allocation failure 1 in matrix()");
   m -= nrl;
   for (i = nrl; i <= nrh; i++)
   {
      m[i] = (float *) malloc((unsigned) (nch - ncl + 1) * sizeof(float));
      if (!m[i])
	 Error("allocation failure 2 in matrix()");
      m[i] -= ncl;
   }
   return (m);
}

/* allocates an double matrix with range [nrl...nrh][ncl...nch]  */
double **dmatrix(int nrl, int nrh, int ncl, int nch)
{
   int i;
   double **m;
   m = (double **) malloc((unsigned) (nrh - nrl + 1) * sizeof(double *));
   if (!m)
      Error("allocation failure 1 in dmatrix()");
   m -= nrl;
   for (i = nrl; i <= nrh; i++)
   {
      m[i] = (double *) malloc((unsigned) (nch - ncl + 1) * sizeof(double));
      if (!m[i])
	 Error("allocation failure 2 in dmatrix()");
      m[i] -= ncl;
   }
   return (m);
}

/* allocates an integer matrix with range [nrl...nrh][ncl...nch]  */
int **imatrix(int nrl, int nrh, int ncl, int nch)
{
   int i;
   int **m;
   m = (int **) malloc((unsigned) (nrh - nrl + 1) * sizeof(int *));
   if (!m)
      Error("allocation failure 1 in imatrix()");
   m -= nrl;
   for (i = nrl; i <= nrh; i++)
   {
      m[i] = (int *) malloc((unsigned) (nch - ncl + 1) * sizeof(int));
      if (!m[i])
	 Error("allocation failure 2 in imatrix()");
      m[i] -= ncl;
   }
   return (m);
}

/* frees a float vector allocated by vector() */
void free_vector(float *v, int nl, int nh)
{
   free((char *) (v + nl));
}

/* frees a int vector allocated by ivector() */
void free_ivector(int *v, int nl, int nh)
{
   free((char *) (v + nl));
}

/* frees a double vector allocated by dvector() */
void free_dvector(double *v, int nl, int nh)
{
   free((char *) (v + nl));
}

/*frees a matrix allocated with matrix. */
void free_matrix(float **m, int nrl, int nrh, int ncl, int nch)
{
   int i;

   for (i = nrh; i >= nrl; i--)
      free((char *) (m[i] + ncl));
   free((char *) (m + nrl));
}

/*frees a matrix allocated with dmatrix. */
void free_dmatrix(double **m, int nrl, int nrh, int ncl, int nch)
{
   int i;

   for (i = nrh; i >= nrl; i--)
      free((char *) (m[i] + ncl));
   free((char *) (m + nrl));
}

/*frees a matrix allocated with imatrix. */
void free_imatrix(int **m, int nrl, int nrh, int ncl, int nch)
{
   int i;

   for (i = nrh; i >= nrl; i--)
      free((char *) (m[i] + ncl));
   free((char *) (m + nrl));
}

/* LU decomposition function */
void ludcmp(float **a, int n, int *indx, double *d)
{
   int i, imax, j, k;
   double big, dum, sum, temp;
   double *vv;

   vv = dvector(1, n);
   *d = 1.0;
   for (i = 1; i <= n; i++)
   {
      big = 0.0;
      for (j = 1; j <= n; j++)
	 if ((temp = fabs(a[i][j])) > big)
	    big = temp;
      if (big == 0.0)
	 Error("Singular matrix in routine LUDCMP");
      vv[i] = 1.0 / big;
   }
   for (j = 1; j <= n; j++)
   {
      for (i = 1; i < j; i++)
      {
	 sum = a[i][j];
	 for (k = 1; k < i; k++)
	    sum -= a[i][k] * a[k][j];
	 a[i][j] = sum;
      }
      big = 0.0;
      for (i = j; i <= n; i++)
      {
	 sum = a[i][j];
	 for (k = 1; k < j; k++)
	    sum -= a[i][k] * a[k][j];
	 a[i][j] = sum;
	 if ((dum = vv[i] * fabs(sum)) >= big)
	 {
	    big = dum;
	    imax = i;
	 }
      }
      if (j != imax)
      {
	 for (k = 1; k <= n; k++)
	 {
	    dum = a[imax][k];
	    a[imax][k] = a[j][k];
	    a[j][k] = dum;
	 }
	 *d = -(*d);
	 vv[imax] = vv[j];
      }
      indx[j] = imax;
      if (a[j][j] == 0.0)
	 a[j][j] = TINY;
      if (j != n)
      {
	 dum = 1.0 / (a[j][j]);
	 for (i = j + 1; i <= n; i++)
	    a[i][j] *= dum;
      }
   }
   free_dvector(vv, 1, n);
}

/* LU back substitution function */
void lubksb(float **a, int n, int *indx, double b[])
{
   int i, ii = 0, ip, j;
   double sum;

   for (i = 1; i <= n; i++)
   {
      ip = indx[i];
      sum = b[ip];
      b[ip] = b[i];
      if (ii)
	 for (j = ii; j <= i - 1; j++)
	    sum -= a[i][j] * b[j];
      else if (sum)
	 ii = i;
      b[i] = sum;
   }

   for (i = n; i >= 1; i--)
   {
      sum = b[i];
      for (j = i + 1; j <= n; j++)
	 sum -= a[i][j] * b[j];
      b[i] = sum / a[i][i];
   }
}

/* Invert matrix function, uses LU decomposition functions */
void invert_matrix(float **a, float **y, int n)
{
   int i, j, *indx;
   double d, *col;

   indx = ivector(1, n);
   col = dvector(1, n);
   ludcmp(a, n, indx, &d);

   for (j = 1; j <= n; j++)
   {
      for (i = 1; i <= n; i++)
	 col[i] = 0.0;
      col[j] = 1.0;
      lubksb(a, n, indx, col);
      for (i = 1; i <= n; i++)
	 y[i][j] = col[i];
   }
}
