/*---------------------------------------------------------------------------*/
/* Program:  metz.c                                                          */
/*                                                                           */
/* Purpose:  This program performs Metz filtering. Images are automatically  */
/*           converted to COMPLEX type, filtered, and converted back to the  */
/*           original type.                                                  */
/*                                                                           */
/* Author:   Chunyen Liu                                                     */
/*                                                                           */
/* Date:     Nov 14, 1995                                                    */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
double exp();
double pow();
float StdDev = 1;
int Power = 10000;
int Debug = FALSE;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs 1D Metz filtering.                        */
/*---------------------------------------------------------------------------*/
void Metz1D(COMPLEX_TYPE * Data, int Xdim)
{
   int x, x1, halfx;
   double numerator, denominator, H, Q;

   /* Convert to frequency domain */
   if (fft_1d(Data, Xdim, -1) == INVALID)
      Error("Could not perform FFT");

   /* Prepare to filter */
   denominator = -2.0 * StdDev * StdDev;
   halfx = Xdim / 2;

   /* Loop over X axis */
   for (x = 0; x < Xdim; x++)
   {
      if (x < halfx)
	 x1 = x;
      else
	 x1 = Xdim - x;

      /* Calculate value of unit height Gaussian H */
      numerator = x1 * x1;
      H = exp(numerator / denominator);

      /* Do pointwise multiplication */
      if (H == 0)
	 Q = 0;
      else
	 Q = (1 - pow((double) (1 - H * H), (double) (Power + 1))) / H;
      if (Debug && x < 10)
	 printf("x=%d H=%f Q=%f\n", x, H, Q);
      Data[x].re *= Q;
      Data[x].im *= Q;
   }

   /* Convert to spatial domain */
   if (fft_1d(Data, Xdim, 1) == INVALID)
      Error("Could not perform FFT");
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs 2D Metz filtering.                        */
/*---------------------------------------------------------------------------*/
void Metz2D(COMPLEX_TYPE ** Data, int Xdim, int Ydim)
{
   int x, y, x1, y1, halfx, halfy;
   double numerator, denominator, H, Q;

   /* Convert to frequency domain */
   if (fft_2d(&(Data[0][0]), Xdim, Ydim, -1) == INVALID)
      Error("Could not perform FFT");

   /* Prepare to filter */
   denominator = -2.0 * StdDev * StdDev;
   halfx = Xdim / 2;
   halfy = Ydim / 2;

   /* Loop over Y axis */
   for (y = 0; y < Ydim; y++)
   {
      if (y < halfy)
	 y1 = y;
      else
	 y1 = Ydim - y;

      /* Loop over X axis */
      for (x = 0; x < Xdim; x++)
      {
	 if (x < halfx)
	    x1 = x;
	 else
	    x1 = Xdim - x;

	 /* Calculate value of unit height Gaussian H */
	 numerator = x1 * x1 + y1 * y1;
	 H = exp(numerator / denominator);

	 /* Do pointwise multiplication */
	 if (H == 0)
	    Q = 0;
	 else
	    Q = (1 - pow((double) (1 - H * H), (double) (Power + 1))) / H;
	 if (Debug && x < 5 && y < 5)
	    printf("x=%d y=%d H=%f Q=%f\n", x, y, H, Q);
	 Data[y][x].re *= Q;
	 Data[y][x].im *= Q;
      }
   }

   /* Convert to spatial domain */
   if (fft_2d(&(Data[0][0]), Xdim, Ydim, 1) == INVALID)
      Error("Could not perform inverse FFT");
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs 3D Metz filtering.                        */
/*---------------------------------------------------------------------------*/
void Metz3D(COMPLEX_TYPE *** Data, int Xdim, int Ydim, int Zdim)
{
   int x, y, z, x1, y1, z1, halfx, halfy, halfz;
   double numerator, denominator, H, Q;

   /* Convert to frequency domain */
   if (fft_3d(&(Data[0][0][0]), Xdim, Ydim, Zdim, -1) == INVALID)
      Error("Could not perform FFT");

   /* Prepare to filter */
   denominator = -2.0 * StdDev * StdDev;
   halfx = Xdim / 2;
   halfy = Ydim / 2;
   halfz = Zdim / 2;

   /* Loop over Z axis */
   for (z = 0; z < Zdim; z++)
   {
      if (z < halfz)
	 z1 = z;
      else
	 z1 = Zdim - z;

      /* Loop over Y axis */
      for (y = 0; y < Ydim; y++)
      {
	 if (y < halfy)
	    y1 = y;
	 else
	    y1 = Ydim - y;

	 /* Loop over X axis */
	 for (x = 0; x < Xdim; x++)
	 {
	    if (x < halfx)
	       x1 = x;
	    else
	       x1 = Xdim - x;

	    /* Calculate value of unit height Gaussian H */
	    numerator = x1 * x1 + y1 * y1 + z1 * z1;
	    H = exp(numerator / denominator);

	    /* Do pointwise multiplication */
	    if (H == 0)
	       Q = 0;
	    else
	       Q = (1 - pow((double) (1 - H * H), (double) (Power + 1))) / H;
	    if (Debug && x < 5 && y < 5 && z < 5)
	       printf("x=%d y=%d z=%d H=%f Q=%f\n", x, y, z, H, Q);
	    Data[z][y][x].re *= Q;
	    Data[z][y][x].im *= Q;
	 }
      }
   }

   /* Convert to spatial domain */
   if (fft_3d(&(Data[0][0][0]), Xdim, Ydim, Zdim, 1) == INVALID)
      Error("Could not perform inverse FFT");
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   COMPLEX_TYPE *Data1;
   COMPLEX_TYPE **Data2;
   COMPLEX_TYPE ***Data3;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int i = 0;
   int Spatial = TRUE;

   /* Interpret program options */
   printf("METZ Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 's':
      case 'S':
	 Spatial = TRUE;
	 if (sscanf(argv[++i], "%f", &StdDev) == 0)
	    Error("Could not get spatial width");
	 break;
      case 'f':
      case 'F':
	 Spatial = FALSE;
	 if (sscanf(argv[++i], "%f", &StdDev) == 0)
	    Error("Could not get frequency width");
	 break;
      case 'k':
	 if (sscanf(argv[++i], "%d", &Power) == 0)
	    Error("Could not get power k");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: metz [options] infile outfile\n");
      fprintf(stderr, "            [-d]      Print debugging information\n");
      fprintf(stderr, "            [-s #]    Width of unit height spatial Gaussian H (1)\n");
      fprintf(stderr, "            [-f #]    Width of unit height frequency Gaussian H (1)\n");
      fprintf(stderr, "            [-k #]    Power k (10000)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image and create output image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);

   /* Convert spatial width into frequency width */
   /* Xdim is used to convert spatial x into frequency u */
   if (Spatial == TRUE)
      StdDev = Xdim / (2.0 * M_PI * StdDev);
   if (Debug == TRUE)
      printf("StdDev = %f\n", StdDev);

   /* Handle 1D images */
   if (DimCnt == 1)
   {
      Data1 = (COMPLEX_TYPE *) im_alloc1D(Image1, COMPLEX);
      im_read(Image1, COMPLEX, (char *) Data1);
      Metz1D(Data1, Xdim);
      im_write(Image2, COMPLEX, (char *) Data1);
      im_free1D((char *) Data1);
   }

   /* Handle 2D images */
   else if (DimCnt == 2)
   {
      Data2 = (COMPLEX_TYPE **) im_alloc2D(Image1, COMPLEX);
      im_read(Image1, COMPLEX, (char *) &(Data2[0][0]));
      Metz2D(Data2, Xdim, Ydim);
      im_write(Image2, COMPLEX, (char *) &(Data2[0][0]));
      im_free2D((char **) Data2);
   }

   /* Handle 3D images */
   else if (DimCnt == 3)
   {
      Data3 = (COMPLEX_TYPE ***) im_alloc3D(Image1, COMPLEX);
      im_read(Image1, COMPLEX, (char *) &(Data3[0][0][0]));
      Metz3D(Data3, Xdim, Ydim, Zdim);
      im_write(Image2, COMPLEX, (char *) &(Data3[0][0][0]));
      im_free3D((char ***) Data3);
   }

   return (0);
}
