/*---------------------------------------------------------------------------*/
/* Program:  morph.c                                                         */
/*                                                                           */
/* Purpose:  This program implements the image morphology operations         */
/*           OPEN, CLOSE, ERODE, and DILATE.  A command line argument        */
/*           specifies the order of these operations to be applied.          */
/*           For example, when the operations string is "oc" an OPEN         */
/*           operation is applied then a CLOSE is applied to the image.      */
/*           Since OPEN = (ERODE, DILATE) and CLOSE = (DILATE, ERODE)        */
/*           the same result could be obtained using "edde" as the           */
/*           operations string.  At present the operations string is         */
/*           limited to 100 characters.                                      */
/*                                                                           */
/* Author:   John Gauch - original program.                                  */
/*           Toni Woodhouse - circular regions.                              */
/*                                                                           */
/* Date:     April 11, 1994                                                  */
/*           July 27, 1995                                                   */
/*           June 18, 1996                                                   */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994-1996               */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Program variables */
int Debug = FALSE;
int Radius = 1;
int ChangeThresh = 100;

FLOAT_TYPE **Data1;
FLOAT_TYPE **Data2;
char Operations[100];
int Xdim, Ydim, Zdim;
double sqrt();


/*---------------------------------------------------------------------------*/
/* Purpose:  This routine does one iteration of DILATE.                      */
/*---------------------------------------------------------------------------*/
void Dilate()
{
   FLOAT_TYPE **swap;
   int x, y, dx, dy;
   float max, dist2;
   float radius2 = Radius * Radius;

   /* Process entire image */
   if (Debug == TRUE)
      printf("Dilate\n");

   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {

	 /* Look at neighborhood */
	 max = Data1[y][x];
	 for (dy = -Radius; dy <= Radius; dy++)
	    for (dx = -Radius; dx <= Radius; dx++)
	    {
	       dist2 = dx * dx + dy * dy;
	       if ((dist2 <= radius2) && (x + dx > 0) && (x + dx < Xdim) &&
		   (y + dy > 0) && (y + dy < Ydim) && (Data1[y + dy][x + dx] > max))
		  max = Data1[y + dy][x + dx];
	    }

	 /* Update pixel value */
	 if ((max - Data1[y][x]) < ChangeThresh)
	    Data2[y][x] = max;
	 else
	    Data2[y][x] = Data1[y][x] + ChangeThresh;
      }

   /* Swap image1 and image2 for next iteration */
   swap = Data1;
   Data1 = Data2;
   Data2 = swap;
}


/*---------------------------------------------------------------------------*/
/* Purpose:  This routine does one iteration of ERODE.                       */
/*---------------------------------------------------------------------------*/
void Erode()
{
   FLOAT_TYPE **swap;
   int x, y, dx, dy;
   float min, dist2;
   float radius2 = Radius * Radius;

   /* Process entire image */
   if (Debug == TRUE)
      printf("Erode\n");

   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {

	 /* Look at neighborhood */
	 min = Data1[y][x];
	 for (dy = -Radius; dy <= Radius; dy++)
	    for (dx = -Radius; dx <= Radius; dx++)
	    {
	       dist2 = dx * dx + dy * dy;
	       if ((dist2 <= radius2) && (x + dx > 0) && (x + dx < Xdim) &&
		   (y + dy > 0) && (y + dy < Ydim) && (Data1[y + dy][x + dx] < min))
		  min = Data1[y + dy][x + dx];
	    }

	 /* Update pixel value */
	 if ((Data1[y][x] - min) < ChangeThresh)
	    Data2[y][x] = min;
	 else
	    Data2[y][x] = Data1[y][x] - ChangeThresh;
      }

   /* Swap image1 and image2 for next iteration */
   swap = Data1;
   Data1 = Data2;
   Data2 = swap;
}


/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   int PixType, DimCnt;

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("MORPH Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'c':
	 if (sscanf(argv[++i], "%d", &ChangeThresh) == 0)
	    Error("Can not get maximum change threshold");
	 break;
      case 'r':
	 if (sscanf(argv[++i], "%d", &Radius) == 0)
	    Error("Can not get radius");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr, "Usage: morph [options] operations infile outfile\n");
      fprintf(stderr, "       [-d]       Print debugging information\n");
      fprintf(stderr, "       [-r #]     Radius of morphology element (1)\n");
      fprintf(stderr, "       [-c #]     Intensity change threshold (100)\n");
      fprintf(stderr, "         e        Operation Erode\n");
      fprintf(stderr, "         d        Operation Dilate\n");
      fprintf(stderr, "         o        Operation Open  = (Erode, Dilate)\n");
      fprintf(stderr, "         c        Operation Close = (Dilate, Erode)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Operations) == 0)
      Error("Could not get morphology operations");
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);

   /* Process operations string one character at a time */
   for (i = 0; i < strlen(Operations); i++)
   {
      switch (Operations[i])
      {
      case 'e':
	 Erode();
	 break;
      case 'd':
	 Dilate();
	 break;
      case 'o':
	 Erode();
	 Dilate();
	 break;
      case 'c':
	 Dilate();
	 Erode();
	 break;
      default:
	 Error("Invalid operation encountered");
	 break;
      }
   }

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data1[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);

   return (0);
}
