/*---------------------------------------------------------------------------*/
/* Program:  needle.c                                                        */
/*                                                                           */
/* Purpose:  This program takes two input optic flow (U,V) vectors and       */
/*           outputs the needle diagram in KUIM format.  The larger dots in  */
/*           the diagram are the arrowheads of vectors.                      */
/*                                                                           */
/* Author:   John Gauch / Chunyen Liu                                        */
/*                                                                           */
/* Date:     March 14, 1996                                                  */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1996                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define INTENSITY 250
#define TOLERENCE 0.001

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   FLOAT_TYPE **Data1;
   FLOAT_TYPE **Data2;
   BYTE_TYPE **Data3;
   int PixType, Xdim, Ydim, Zdim, DimCnt;
   int Xdim2, Ydim2, Zdim2;

   /* Program variables */
   int Debug = FALSE;
   int inverse = 0;
   float dU, dV, scale = 2.0;
   int i = 0, x, y, newX, newY, length, interval = 10;

   /* Interpret program options */
   printf("NEEDLE Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      case 'r':
	 inverse = 1;
	 break;
      case 's':
	 if (sscanf(argv[++i], "%f", &scale) == 0)
	    Error("Could not get scale argument");
	 break;
      case 'i':
	 if (sscanf(argv[++i], "%d", &interval) == 0)
	    Error("Could not get interval argument");
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr, "Usage: needle [options] infileU infileV outfile\n");
      fprintf(stderr, "              [-s #]    Scaling factor for both vectors (2.0)\n");
      fprintf(stderr, "              [-i #]    Interval size for display (10)\n");
      fprintf(stderr, "              [-r]      Inverse for printing\n");
      fprintf(stderr, "              [-d]      Print debugging information\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name3) == 0)
      Error("Could not get output file name");

   /* Read input image1 */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (PixType == COLOR || PixType == PSEUDO || PixType == JPEG_RGB)
      Error("Can not handle COLOR images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Read input image2 */
   Image2 = im_open(Name2, &PixType, &Xdim2, &Ydim2, &Zdim2, &DimCnt);
   if ((Xdim != Xdim2) || (Ydim != Ydim2) || (Zdim != Zdim2))
      Error("Image dimensions do not match");
   if (PixType == COLOR || PixType == PSEUDO || PixType == JPEG_RGB)
      Error("Can not handle COLOR images");
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   im_read(Image2, FLOAT, (char *) &(Data2[0][0]));

   /* Create output image */
   Image3 = im_create(Name3, BYTE, Xdim, Ydim, Zdim);
   Data3 = (BYTE_TYPE **) im_alloc2D(Image3, BYTE);

   /* Initialize the needle diagram */
   for (x = 0; x < Xdim; x++)
      for (y = 0; y < Ydim; y++)
	 Data3[y][x] = inverse * INTENSITY;

   /* Interpolate and fill in all vectors  */
   for (x = 0; x < Xdim; x += interval)
      for (y = 0; y < Ydim; y += interval)
      {
	 dU = scale * Data1[y][x];
	 dV = scale * Data2[y][x];
	 Data3[y][x] = (1 - inverse) * INTENSITY;
	 length = (int) sqrt((double) (dU * dU + dV * dV));
	 if (length > TOLERENCE)
	    for (i = length; i > 0; i--)
	    {
	       newX = (int) (x + dU * i / length);
	       newY = (int) (y + dV * i / length);
	       if ((newX >= 0) && (newX < Xdim) && (newY >= 0) && (newY < Ydim))
		  Data3[newY][newX] = (1 - inverse) * INTENSITY;
	       /* Draw the end point of the vector */
	       if (i == length)
	       {
		  if ((newX + 1 >= 0) && (newX + 1 < Xdim) && (newY >= 0) && (newY < Ydim))
		     Data3[newY][newX + 1] = (1 - inverse) * INTENSITY;
		  if ((newX - 1 >= 0) && (newX - 1 < Xdim) && (newY >= 0) && (newY < Ydim))
		     Data3[newY][newX - 1] = (1 - inverse) * INTENSITY;
		  if ((newX >= 0) && (newX < Xdim) && (newY + 1 >= 0) && (newY + 1 < Ydim))
		     Data3[newY + 1][newX] = (1 - inverse) * INTENSITY;
		  if ((newX >= 0) && (newX < Xdim) && (newY - 1 >= 0) && (newY - 1 < Ydim))
		     Data3[newY - 1][newX] = (1 - inverse) * INTENSITY;
	       }
	    }
      }

   /* Write information to output image */
   im_write(Image3, BYTE, (char *) &(Data3[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   im_free2D((char **) Data3);
   return (0);
}
