/*---------------------------------------------------------------------------*/
/* Program:  plot_contour.c                                                  */
/*                                                                           */
/* Purpose:  This program plots a contour on top of an image.                */
/*           These points are expected to be in an ascii file with           */
/*           (x,y) coordinates one per line.  Each contour starts            */
/*           with "START CONTOUR" and ends with "-1 -1".                     */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     May 2, 1994.                                                    */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Global variables */
FLOAT_TYPE **Data1;
FLOAT_TYPE **Data2;
int Intensity = 255;
int Thick = FALSE;

/* Function prototype */
void PlotLine(int x1, int y1, int x2, int y2);

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   int PixType, Udim, Vdim, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int i = 0, x, y;

   /* Interpret program options */
   printf("PLOT_CONTOUR Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'i':
	 if (sscanf(argv[++i], "%d", &Intensity) == 0)
	    Error("Could not get integer argument");
	 break;
      case 't':
	 Thick = TRUE;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: plot_contour [options] confile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-t]    Print thick lines on image\n");
      fprintf(stderr, "       [-i #]  Intensity of plotted points (255)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input contour */
   Image1 = im_open(Name1, &PixType, &Udim, &Vdim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can only process 2D images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Xdim = Ydim = 128;
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 Data2[y][x] = 0.0;

   /* Plot lines of contour */
   for (i = 1; i < Udim; i++)
      PlotLine((int) Data1[0][i], (int) Data1[1][i],
	       (int) Data1[0][i - 1], (int) Data1[1][i - 1]);
   PlotLine((int) Data1[0][0], (int) Data1[1][0],
	    (int) Data1[0][Udim - 1], (int) Data1[1][Udim - 1]);

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   return (0);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine plots a given line segment.                        */
/*---------------------------------------------------------------------------*/
void PlotLine(int x1, int y1, int x2, int y2)
{
   int dx, dy, length, i, x, y;

   dx = x2 - x1;
   dy = y2 - y1;
   length = (int) sqrt((double) (dx * dx + dy * dy));
   for (i = 0; i < length; i++)
   {
      x = x1 + dx * i / length;
      y = y1 + dy * i / length;
      Data2[y][x] = Intensity;
      if (Thick == TRUE)
      {
	 Data2[y + 1][x] = Intensity;
	 Data2[y][x + 1] = Intensity;
	 Data2[y + 1][x + 1] = Intensity;
      }
   }
}
