/*---------------------------------------------------------------------------*/
/* Program:  pre_warp.c                                                      */
/*                                                                           */
/* Purpose:  This program reads in two body contours in KUIM format and      */
/*           calculates a set of control points for warping from one to      */
/*           the other.  The geometric center of both contours are first     */
/*           aligned and the N points are chosen along evenly spaced         */
/*           directions.                                                     */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     June, 1 1995                                                    */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   FLOAT_TYPE **Data1;
   FLOAT_TYPE **Data2;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   FILE *fd;
   int Debug = FALSE;
   int i = 0, j;
   int PtCnt1, PtCnt2;
   float Xmean1, Ymean1;
   float Xmean2, Ymean2;

   /* Interpret program options */
   printf("PRE_WARP Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr, "Usage: ore_warp [options] contour1 contour2 outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-n #]  Integer program argument\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name3) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Read input image */
   Image2 = im_open(Name2, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   im_read(Image2, FLOAT, (char *) &(Data2[0][0]));

   /* Calculate center of first contour */
   PtCnt1 = (int) (0.5 + Data1[0][Xdim - 1]);
   if (Debug == TRUE)
      printf("ptcnt %d\n", PtCnt1);
   Xmean1 = Ymean1 = 0;
   for (i = 0; i < PtCnt1; i++)
   {
      Xmean1 += Data1[0][i];
      Ymean1 += Data1[1][i];
   }
   if (PtCnt1 > 0)
   {
      Xmean1 = Xmean1 / PtCnt1;
      Ymean1 = Ymean1 / PtCnt1;
   }
   if (Debug == TRUE)
      printf("center %f %f\n", Xmean1, Ymean1);

   /* Calculate center of second contour */
   PtCnt2 = (int) (0.5 + Data2[0][Xdim - 1]);
   if (Debug == TRUE)
      printf("ptcnt %d\n", PtCnt2);
   Xmean2 = Ymean2 = 0;
   for (j = 0; j < PtCnt2; j++)
   {
      Xmean2 += Data2[0][j];
      Ymean2 += Data2[1][j];
   }
   if (PtCnt2 > 0)
   {
      Xmean2 = Xmean2 / PtCnt2;
      Ymean2 = Ymean2 / PtCnt2;
   }
   if (Debug == TRUE)
      printf("center %f %f\n", Xmean2, Ymean2);

   /* Move contours centers to origin */
   for (i = 0; i < PtCnt1; i++)
   {
      Data1[0][i] -= Xmean1;
      Data1[1][i] -= Ymean1;
   }
   for (j = 0; j < PtCnt2; j++)
   {
      Data2[0][j] -= Xmean2;
      Data2[1][j] -= Ymean2;
   }

   /* Create output file */
   if ((fd = fopen(Name3, "w")) == NULL)
      Error("Could not create point file");

   /* Find y=0, x<0 points */
   i = 0;
   j = 0;
   while (Data1[1][i] > 0)
      i = (i + 1) % PtCnt1;
   while (Data1[1][i] < 0)
      i = (i + 1) % PtCnt1;
   while (Data2[1][j] > 0)
      j = (j + 1) % PtCnt2;
   while (Data2[1][j] < 0)
      j = (j + 1) % PtCnt2;
   fprintf(fd, "%f %f %f %f\n", Xmean1 + Data1[0][i], Ymean1 + Data1[1][i],
	   Xmean2 + Data2[0][j], Ymean2 + Data2[1][j]);

   /* Find -y=x, x<0 points */
   while (Data1[1][i] < -Data1[0][i])
      i = (i + 1) % PtCnt1;
   while (Data2[1][j] < -Data2[0][j])
      j = (j + 1) % PtCnt2;
   fprintf(fd, "%f %f %f %f\n", Xmean1 + Data1[0][i], Ymean1 + Data1[1][i],
	   Xmean2 + Data2[0][j], Ymean2 + Data2[1][j]);

   /* Find y>0, x=0 points */
   while (Data1[0][i] < 0)
      i = (i + 1) % PtCnt1;
   while (Data2[0][j] < 0)
      j = (j + 1) % PtCnt2;
   fprintf(fd, "%f %f %f %f\n", Xmean1 + Data1[0][i], Ymean1 + Data1[1][i],
	   Xmean2 + Data2[0][j], Ymean2 + Data2[1][j]);

   /* Find y=x, x>0 points */
   while (Data1[0][i] < Data1[1][i])
      i = (i + 1) % PtCnt1;
   while (Data2[0][j] < Data2[1][j])
      j = (j + 1) % PtCnt2;
   fprintf(fd, "%f %f %f %f\n", Xmean1 + Data1[0][i], Ymean1 + Data1[1][i],
	   Xmean2 + Data2[0][j], Ymean2 + Data2[1][j]);

   /* Find y=0, x>0 points */
   while (Data1[1][i] > 0)
      i = (i + 1) % PtCnt1;
   while (Data2[1][j] > 0)
      j = (j + 1) % PtCnt2;
   fprintf(fd, "%f %f %f %f\n", Xmean1 + Data1[0][i], Ymean1 + Data1[1][i],
	   Xmean2 + Data2[0][j], Ymean2 + Data2[1][j]);

   /* Find -y=x, x>0 points */
   while (Data1[1][i] > -Data1[0][i])
      i = (i + 1) % PtCnt1;
   while (Data2[1][j] > -Data2[0][j])
      j = (j + 1) % PtCnt2;
   fprintf(fd, "%f %f %f %f\n", Xmean1 + Data1[0][i], Ymean1 + Data1[1][i],
	   Xmean2 + Data2[0][j], Ymean2 + Data2[1][j]);

   /* Find y<0, x=0 points */
   while (Data1[0][i] > 0)
      i = (i + 1) % PtCnt1;
   while (Data2[0][j] > 0)
      j = (j + 1) % PtCnt2;
   fprintf(fd, "%f %f %f %f\n", Xmean1 + Data1[0][i], Ymean1 + Data1[1][i],
	   Xmean2 + Data2[0][j], Ymean2 + Data2[1][j]);

   /* Find y=x, x<0 points */
   while (Data1[0][i] > Data1[1][i])
      i = (i + 1) % PtCnt1;
   while (Data2[0][j] > Data2[1][j])
      j = (j + 1) % PtCnt2;
   fprintf(fd, "%f %f %f %f\n", Xmean1 + Data1[0][i], Ymean1 + Data1[1][i],
	   Xmean2 + Data2[0][j], Ymean2 + Data2[1][j]);

   /* Close output file */
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   fclose(fd);
   return (0);
}
