/*---------------------------------------------------------------------------*/
/* Program:  predict.c                                                       */
/*                                                                           */
/* Purpose:  This program implements several types of pixel prediction       */
/*           schemes which are useful for lowering image entropy prior       */
/*           to encoding.  The result should be higher image compression     */
/*           rates for most images.                                          */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     November 13, 1995                                               */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   SHORT_TYPE **Data1;
   SHORT_TYPE **Data2;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Encode = TRUE;
   int Predict = 1;
   int i = 0, x, y;
   int A, B, C;

   /* Interpret program options */
   printf("PREDICT Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'e':
	 Encode = TRUE;
	 if (sscanf(argv[++i], "%d", &Predict) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'd':
	 Encode = FALSE;
	 if (sscanf(argv[++i], "%d", &Predict) == 0)
	    Error("Could not get integer argument");
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: predict [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-e #]  Calculate and store predictions\n");
      fprintf(stderr, "       [-d #]  Recover image from predictions\n");
      fprintf(stderr, "Prediction methods:\n");
      fprintf(stderr, "       1 => x=A\n");
      fprintf(stderr, "       2 => x=B\n");
      fprintf(stderr, "       3 => x=C\n");
      fprintf(stderr, "       4 => x=A+B-C\n");
      fprintf(stderr, "       5 => x=A+(B-C)/2\n");
      fprintf(stderr, "       6 => x=B+(A-C)/2\n");
      fprintf(stderr, "       7 => x=(A+B)/2\n");
      fprintf(stderr, "       8 => x=(A+B+C)/3\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);
   im_read(Image1, SHORT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);

   /* Generate predicted error image */
   if (Encode == TRUE)
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    /* Get neighborhood values */
	    if (x > 0)
	       A = Data1[y][x - 1];
	    else
	       A = 0;
	    if (y > 0)
	       B = Data1[y - 1][x];
	    else
	       B = 0;
	    if (x > 0 && y > 0)
	       C = Data1[y - 1][x - 1];
	    else
	       C = 0;

	    /* Generate predicted error image */
	    switch (Predict)
	    {
	    case 1:
	       Data2[y][x] = Data1[y][x] - A;
	       break;
	    case 2:
	       Data2[y][x] = Data1[y][x] - B;
	       break;
	    case 3:
	       Data2[y][x] = Data1[y][x] - C;
	       break;
	    case 4:
	       Data2[y][x] = Data1[y][x] - A - B + C;
	       break;
	    case 5:
	       Data2[y][x] = Data1[y][x] - A - (B - C) / 2;
	       break;
	    case 6:
	       Data2[y][x] = Data1[y][x] - B - (A - C) / 2;
	       break;
	    case 7:
	       Data2[y][x] = Data1[y][x] - (A + B) / 2;
	       break;
	    case 8:
	       Data2[y][x] = Data1[y][x] - (A + B + C) / 3;
	       break;
	    default:
	       Data2[y][x] = Data1[y][x];
	       break;
	    }
	 }

   /* Recover predicted image from errors */
   if (Encode == FALSE)
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    /* Get neighborhood values */
	    if (x > 0)
	       A = Data2[y][x - 1];
	    else
	       A = 0;
	    if (y > 0)
	       B = Data2[y - 1][x];
	    else
	       B = 0;
	    if (x > 0 && y > 0)
	       C = Data2[y - 1][x - 1];
	    else
	       C = 0;

	    /* Recover predicted image from errors */
	    switch (Predict)
	    {
	    case 1:
	       Data2[y][x] = Data1[y][x] + A;
	       break;
	    case 2:
	       Data2[y][x] = Data1[y][x] + B;
	       break;
	    case 3:
	       Data2[y][x] = Data1[y][x] + C;
	       break;
	    case 4:
	       Data2[y][x] = Data1[y][x] + A + B - C;
	       break;
	    case 5:
	       Data2[y][x] = Data1[y][x] + A + (B - C) / 2;
	       break;
	    case 6:
	       Data2[y][x] = Data1[y][x] + B + (A - C) / 2;
	       break;
	    case 7:
	       Data2[y][x] = Data1[y][x] + (A + B) / 2;
	       break;
	    case 8:
	       Data2[y][x] = Data1[y][x] + (A + B + C) / 3;
	       break;
	    default:
	       Data2[y][x] = Data1[y][x];
	       break;
	    }
	 }

   /* Write information to output image */
   im_write(Image2, SHORT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   return (0);
}
