/*---------------------------------------------------------------------------*/
/* Program:  project.c                                                       */
/*                                                                           */
/* Purpose:  This program demonstrates the first stage of computed           */
/*           tomography.  The input image is parallel projected from         */
/*           a number of angles to obtain a projection image.  The           */
/*           programs FILTER_PROJECT and BACK_PROJECT can be used to         */
/*           complete the reconstruction process.                            */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     May 9, 1994                                                     */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define ROUND(r) ((r) >= 0.0) ? ((int)(r + 0.5)) : ((int)(r - 0.5))

/* Global variables */
FLOAT_TYPE **Data1;
FLOAT_TYPE **Data2;
int Xdim, Ydim, Zdim;
int Debug = FALSE;
int Fast = FALSE;
int NumBuckets;
int NumProject = 200;
double sin();
double cos();

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs parallel projection.                      */
/*---------------------------------------------------------------------------*/
void Project()
{
   int x, y, a, b, Xcenter, Ycenter, Ileft, Iright;
   float val, Position, Aleft, Aright, *SinTab, *CosTab;

   /* Initialize sin and cos lookup tables */
   SinTab = (float *) malloc((unsigned) sizeof(float) * NumProject);
   CosTab = (float *) malloc((unsigned) sizeof(float) * NumProject);
   for (a = 0; a < NumProject; a++)
   {
      SinTab[a] = (float) sin((double) (a * M_PI / NumProject));
      CosTab[a] = (float) cos((double) (a * M_PI / NumProject));
   }

   /* Initialize projections to zero */
   for (a = 0; a < NumProject; a++)
      for (b = 0; b < NumBuckets; b++)
	 Data2[a][b] = 0.0;

   /* Project each pixel in the image */
   Xcenter = Xdim / 2;
   Ycenter = Ydim / 2;
   for (y = -Ycenter; y < Ycenter; y++)
      for (x = -Xcenter; x < Xcenter; x++)
      {
	 val = Data1[y + Ycenter][x + Xcenter];
	 for (a = 0; a < NumProject; a++)
	 {
	    /* Fast inaccurate projection */
	    Position = x * CosTab[a] + y * SinTab[a];
	    if (Fast == TRUE)
	    {
	       /* Calculate coordinates of nearest bucket */
	       b = ROUND(Position);
	       if (b < 0)
		  b += NumBuckets;
	       Data2[a][b] += val;
	    }

	    /* Slower more accurate projection */
	    else
	    {
	       /* Calculate coordinates of left and right pixels */
	       if (Position >= 0.0)
	       {
		  Ileft = (int) Position;
		  Iright = Ileft + 1;
	       }
	       else
	       {
		  Iright = (int) Position;
		  Ileft = Iright - 1;
	       }

	       /* Calculate contribution to left and right pixels */
	       Aleft = Iright - Position;
	       Aright = Position - Ileft;
	       if (Ileft < 0)
		  Ileft += NumBuckets;
	       if (Iright < 0)
		  Iright += NumBuckets;
	       Data2[a][Ileft] += (val * Aleft);
	       Data2[a][Iright] += (val * Aright);
	    }
	 }
      }

   /* Fix 45 and 135 projections for fast method */
   if (Fast == TRUE)
      for (b = 0; b < NumBuckets; b++)
      {
	 Data2[NumProject / 4][b] = Data2[1 + NumProject / 4][b];
	 Data2[3 * NumProject / 4][b] = Data2[1 + 3 * NumProject / 4][b];
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   int PixType, DimCnt;

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("PROJECT Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'p':
	 if (sscanf(argv[++i], "%d", &NumProject) == 0)
	    Error("Could not get NumProject argument");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      case 'f':
	 Fast = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: project [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-f]    Fast but inaccurate projections\n");
      fprintf(stderr, "       [-p #]  Number of parallel projections (200)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   NumBuckets = Xdim * 2;
   Image2 = im_create(Name2, FLOAT, NumBuckets, NumProject, 1);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);

   /* Manipulate input image data */
   Project();

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   return (0);
}
