/*--------------------------------------------------------------------------*/
/* Program:  proto_fft.c                                                    */
/*                                                                          */
/* Purpose:  This program demonstrates filtering in the frequency           */
/*           domain.  Images are automatically converted to COMPLEX type    */
/*           to be filtered and converted back to FLOAT type to be output.  */
/*                                                                          */
/* Author:   John Gauch                                                     */
/*                                                                          */
/* Date:     September 25, 1995                                             */
/*                                                                          */
/* Note:     Copyright (C) The University of Kansas, 1995                   */
/*--------------------------------------------------------------------------*/
#include <IM.h>

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   COMPLEX_TYPE **Data1;
   COMPLEX_TYPE **Data2;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   float Frequency = 30;
   int Debug = FALSE;
   int i = 0, x, y;

   /* Interpret program options */
   printf("PROTO_FFT Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'f':
	 if (sscanf(argv[++i], "%f", &Frequency) == 0)
	    Error("Could not get Frequency argument");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: proto_fft [options] infile outfile\n");
      fprintf(stderr, "       [-f #]  Frequency argument (default = 30)\n");
      return (1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (COMPLEX_TYPE **) im_alloc2D(Image1, COMPLEX);
   im_read(Image1, COMPLEX, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (COMPLEX_TYPE **) im_alloc2D(Image2, COMPLEX);

   /* Convert image to frequency domain */
   if (fft_2d(&(Data1[0][0]), Xdim, Ydim, -1) == INVALID)
      Error("Could not perform forward FFT");

   /* Perform filtering */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Data2[y][x].re = Data1[y][x].re;
	 Data2[y][x].im = Data1[y][x].im;
      }

   /* Convert image back to spatial domain */
   if (fft_2d(&(Data2[0][0]), Xdim, Ydim, 1) == INVALID)
      Error("Could not perform inverse FFT");

   /* Write information to output image */
   im_write(Image2, COMPLEX, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   return (0);
}
