/*---------------------------------------------------------------------------*/
/* Program:  rl_decode.c                                                     */
/*                                                                           */
/* Purpose:  This program generates a KUIM image by performing run length    */
/*           decoding of a byte file with a number of (length,value)         */
/*           pairs representing each scan line in the image. For binary      */
/*           images, we assume that each scan line alternates from 0 to 1    */
/*           starting with a 0 in the first location. A small header is      */
/*           read from the start of the file to get the xdim,ydim,zdim       */
/*           of the output image.                                            */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     March 29, 1995                                                  */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image;
   BYTE_TYPE *Data;
   int Xdim, Ydim, Zdim;

   /* Program variables */
   int Debug = FALSE;
   int i = 0;
   int Index;
   short Header[4];
   BYTE_TYPE Pair[2];
   FILE *Fd;

   /* Interpret program options */
   printf("RL_DECODE Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: rl_decode [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read encoded image header */
   if ((Fd = fopen(Name1, "r")) == NULL)
      Error("Could not open input file");
   fread(Header, 1, sizeof(Header), Fd);
   if (Header[0] != 66)
      Error("Incorrect input file format");
   Xdim = Header[1];
   Ydim = Header[2];
   Zdim = Header[3];

   /* Create output image */
   Image = im_create(Name2, BYTE, Xdim, Ydim, Zdim);
   Data = (BYTE_TYPE *) im_alloc2D(Image, BYTE);

   /* Perform run length decoding */
   Index = 0;
   fread(Pair, 1, sizeof(Pair), Fd);
   while (!feof(Fd))
   {
      while (Pair[0] > 0)
      {
	 Data[Index++] = Pair[1];
	 Pair[0]--;
      }
      fread(Pair, 1, sizeof(Pair), Fd);
   }

   /* Write information to output image */
   im_write(Image, BYTE, (char *) Data);
   im_free1D((char *) Data);
   return (0);
}
