/*---------------------------------------------------------------------------*/
/* Program:  rl_encode.c                                                     */
/*                                                                           */
/* Purpose:  This program performs run length encoding on a KUIM image       */
/*           and generates a byte file with a number of (length,value)       */
/*           pairs representing each scan line in the image. For binary      */
/*           images, we assume that each scan line alternates from 0 to 1    */
/*           starting with a 0 in the first location. A small header is      */
/*           added to the start of the file to give the xdim,ydim,zdim       */
/*           of the original image.                                          */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     March 29, 1995                                                  */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image;
   BYTE_TYPE **Data;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int i = 0, x, y;
   int Count, Value;
   BYTE_TYPE Pair[2];
   short Header[4];
   FILE *Fd;

   /* Interpret program options */
   printf("RL_ENCODE Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: rl_encode [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data = (BYTE_TYPE **) im_alloc2D(Image, BYTE);
   im_read(Image, BYTE, (char *) &(Data[0][0]));

   /* Create output image */
   if ((Fd = fopen(Name2, "w")) == NULL)
      Error("Could not open output file");
   Header[0] = 66;
   Header[1] = Xdim;
   Header[2] = Ydim;
   Header[3] = Zdim;
   fwrite(Header, 1, sizeof(Header), Fd);

   /* Perform run length encoding */
   for (y = 0; y < Ydim; y++)
   {
      Count = 1;
      Value = Data[y][0];
      for (x = 1; x < Xdim; x++)
      {
	 if ((Data[y][x] == Value) && (Count < 255))
	    Count++;
	 else
	 {
	    Pair[0] = Count;
	    Pair[1] = Value;
	    fwrite(Pair, 1, sizeof(Pair), Fd);
	    Count = 1;
	    Value = Data[y][x];
	 }
      }
      Pair[0] = Count;
      Pair[1] = Value;
      fwrite(Pair, 1, sizeof(Pair), Fd);
   }

   /* Close output image */
   fclose(Fd);
   im_free2D((char **) Data);
   return (0);
}
