/*---------------------------------------------------------------------------*/
/* Program:  shade.c                                                         */
/*                                                                           */
/* Purpose:  This program interprets an image as a depth map and             */
/*           generates a shaded image of this surface using a                */
/*           simplified Lambertian lighting model.                           */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     February 22, 1994                                               */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <IM.h>

/* Global variables */
char Name1[50];
char Name2[50];
BYTE_TYPE *Data1;
FLOAT_TYPE **Data2;
FLOAT_TYPE **Data3;
int Debug = FALSE;
char Title[nTITLE];
int Xdim, Ydim, Zdim;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates a shaded image.                         */
/*---------------------------------------------------------------------------*/
void Shade(float Lx, float Ly, float Lz)
{
   int x, y;
   float p, q, length, temp;

   /* Get normalized light vector */
   length = (float) sqrt((double) Lx * Lx + Ly * Ly + Lz * Lz);
   Lx = Lx / length;
   Ly = Ly / length;
   Lz = Lz / length;
   if (Debug == TRUE)
      printf("Lx,Ly,Lz %f %f %f\n", Lx, Ly, Lz);
   sprintf(Title, "%f %f %f", Lx, Ly, Lz);

   /* Calculate shade for each pixel */
   for (y = 1; y < Ydim; y++)
      for (x = 1; x < Xdim; x++)
      {
	 /* Calculate surface normal */
	 p = (float) Data3[y][x] - Data3[y][x - 1];
	 q = (float) Data3[y][x] - Data3[y - 1][x];
	 length = (float) sqrt((double) p * p + q * q + 1);

	 /* Calculate surface illumination */
	 temp = (-Lx * p - Ly * q + Lz);
	 if (temp < 0.0)
	    Data2[y][x] = 0.0;
	 else
	    Data2[y][x] = temp / length;
      }

   /* Handle first row and column */
   for (x = 0; x < Xdim; x++)
      Data2[0][x] = Data2[1][x];
   for (y = 0; y < Ydim; y++)
      Data2[y][0] = Data2[y][1];
}

/*---------------------------------------------------------------------------*/
/* Purpose: This routine displays an image.                                  */
/*---------------------------------------------------------------------------*/
void MyMain()
{
   Display *display;
   char *display_name = NULL;
   int screen_num;
   int depth;
   unsigned long foreground;
   unsigned long background;
   Window window;
   GC gc;
   Visual *visual;
   Colormap old_colormap;
   Colormap new_colormap;
   int i;
   XColor color;
   double Gamma = 0.01;
   int Count, Min, Max;
   XImage *ximage;
   int done = FALSE;
   XEvent event;
   char text[50];
   KeySym key;
   float Lx, Ly, Lz, temp;

   /* Set up window for display */
   if ((display = XOpenDisplay(display_name)) == NULL)
      Error("Can not connect to X server");
   screen_num = DefaultScreen(display);
   depth = DefaultDepth(display, screen_num);
   if (depth != 8)
      Error("Can only display on 8 bit displays");
   background = BlackPixel(display, screen_num);
   foreground = WhitePixel(display, screen_num);
   window = XCreateSimpleWindow(display, DefaultRootWindow(display),
			   200, 200, Xdim, Ydim, 5, foreground, background);
   XStoreName(display, window, Name2);
   XSelectInput(display, window, ButtonPressMask | KeyPressMask | ExposureMask);
   gc = XCreateGC(display, window, 0, 0);
   XSetBackground(display, gc, background);
   XSetForeground(display, gc, foreground);
   visual = XDefaultVisualOfScreen(DefaultScreenOfDisplay(display));
   XMapRaised(display, window);

   /* Set up colormap */
   old_colormap = DefaultColormap(display, screen_num);
   new_colormap = XCreateColormap(display, window, visual, AllocAll);
   for (i = 0; i <= 255; i++)
   {
      color.pixel = i;
      XQueryColor(display, old_colormap, &color);
      color.flags = DoRed | DoGreen | DoBlue;
      XStoreColor(display, new_colormap, &color);
   }
   for (i = 50; i <= 250; i++)
   {
      color.pixel = i;
      color.blue = color.green = color.red = (int) (65535
				    * log((double) (1.0 + Gamma * (i - 50)))
				       / log((double) (1.0 + Gamma * 200)));
      XStoreColor(display, new_colormap, &color);
   }

   /* Copy image to display window */
   XSetWindowColormap(display, window, new_colormap);
   ximage = XCreateImage(display, visual, 8, ZPixmap, 0, Data1, Xdim, Ydim, 8, 0);
   Count = Xdim * Ydim * Zdim;
   for (i = 0; i < Count; i++)
      Data1[i] = (unsigned char) (50 + 200 * Data2[0][i]);

   /* Loop handling events */
   while (done == FALSE)
   {
      XNextEvent(display, &event);
      switch (event.type)
      {
      case Expose:
	 if (event.xexpose.count == 0)
	    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	 break;
      case MappingNotify:
	 XRefreshKeyboardMapping((XMappingEvent *) & event);
	 break;
      case ButtonPress:
	 /* Mouse button clicked, process selected pixel */
	 Lx = (2.0 * event.xbutton.x - Xdim) / Xdim;
	 Ly = (2.0 * event.xbutton.y - Ydim) / Ydim;
	 temp = 1 - Lx * Lx - Ly * Ly;
	 if (temp < 0.0)
	    temp = 0.0;
	 Lz = (float) sqrt(temp);
	 Shade(Lx, Ly, Lz);

	 /* Redraw new shaded image */
	 for (i = 0; i < Count; i++)
	    Data1[i] = (unsigned char) (50 + 200 * Data2[0][i]);
	 XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	 break;
      case KeyPress:
	 /* Exit if 'q' was pressed */
	 i = XLookupString((XKeyEvent *) & event, text, 50, &key, 0);
	 if (i == 1 && text[0] == 'q')
	    done = TRUE;
	 break;
      }
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Image variables */
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   int PixType, Count, DimCnt;

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("SHADE Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: shade [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read FLOAT copy of input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data3 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data3[0][0]));

   /* Make BYTE copy of input image */
   Data1 = (BYTE_TYPE *) im_alloc1D(Image1, BYTE);
   Count = Xdim * Ydim * Zdim;
   im_convert(Image1, Count, (char *) &(Data3[0][0]), Data1, FLOAT, BYTE);

   /* Create FLOAT output image */
   Image2 = im_create(Name2, FLOAT, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   Shade(0.0, 0.0, 1.0);
   MyMain();

   /* Write information to output image */
   im_put_title(Image2, Title);
   im_write(Image2, FLOAT, (char *) &(Data2[0][0]));
   im_free1D((char *) Data1);
   im_free2D((char **) Data2);
   im_free2D((char **) Data3);
   exit(0);
}
