/*---------------------------------------------------------------------------*/
/* Program:  thresh.c                                                        */
/*                                                                           */
/* Purpose:  This program performs statistics based, edge strength based,    */
/*           optimal, or static thresholding of KUIM format images using     */
/*           linear interpolation.                                           */
/*                                                                           */
/* Author:   John Gauch, Edu Metz                                            */
/*                                                                           */
/* Date:     March 23, 1994     - Original program.                          */
/*           September 12, 1996 - Interpolated thresholds added.             */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994-1996               */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Program variables */
int StatsThreshold = FALSE;
int EdgeThreshold = FALSE;
int OptimalThreshold = FALSE;
int StaticThreshold = FALSE;
int Debug = FALSE;
int StoreThreshold = FALSE;
int RegionX = 1;
int RegionY = 1;
float UserThreshold = 0.0;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs statistics based thresholding.            */
/*---------------------------------------------------------------------------*/
float StatThresholding(FLOAT_TYPE ** Data, int Xstart, int Ystart,
		       int Xend, int Yend)
{
   float Threshold;
   float max, min, mean1, mean2;
   int x, y, count1, count2, loop;

   /* Find max and min */
   max = min = Data[0][0];
   for (y = Ystart; y < Yend; y++)
      for (x = Xstart; x < Xend; x++)
	 if (Data[y][x] > max)
	    max = Data[y][x];
	 else if (Data[y][x] < min)
	    min = Data[y][x];
   Threshold = (max - min) / 2;

   /* Find mean of whole distribution */
   mean1 = 0.0;
   for (y = Ystart; y < Yend; y++)
      for (x = Xstart; x < Xend; x++)
	 mean1 += Data[y][x];
   Threshold = mean1 / ((Xstart - Xend) * (Ystart - Yend));

   /* Find mean of two distributions */
   mean1 = mean2 = 0.0;
   count1 = count2 = 1;

   for (y = Ystart; y < Yend; y++)
      for (x = Xstart; x < Xend; x++)
	 if (Data[y][x] < Threshold)
	 {
	    count1++;
	    mean1 += Data[y][x];
	 }
	 else
	 {
	    count2++;
	    mean2 += Data[y][x];
	 }
   Threshold = (mean1 / count1 + mean2 / count2) / 2;

   /* Revise threshold value */
   for (loop = 0; loop < 10; loop++)
   {
      mean1 = mean2 = 0.0;
      count1 = count2 = 1;

      for (y = Ystart; y < Yend; y++)
	 for (x = Xstart; x < Xend; x++)
	    if (Data[y][x] < Threshold)
	    {
	       count1++;
	       mean1 += Data[y][x];
	    }
	    else
	    {
	       count2++;
	       mean2 += Data[y][x];
	    }
      Threshold = (mean1 / count1 + mean2 / count2) / 2;
   }

   return (Threshold);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs edge strength based thresholding.         */
/*---------------------------------------------------------------------------*/
float EdgeThresholding(FLOAT_TYPE ** Data, int Xstart, int Ystart,
		       int Xend, int Yend)
{
   float Threshold, grad, sum_grad = 0.0, sum_edge = 0.0;
   int x, y;

   for (y = Ystart; y < Yend; y++)
      for (x = Xstart + 1; x < Xend; x++)
      {
	 grad = Data[y][x] - Data[y][x - 1];
	 if (grad < 0)
	    grad = -grad;
	 sum_grad += grad;
	 sum_edge += grad * Data[y][x];
      }
   Threshold = sum_edge / sum_grad;

   return (Threshold);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs optimal thresholding.                     */
/*---------------------------------------------------------------------------*/
float OptThresholding(FLOAT_TYPE ** Data, int Xstart, int Ystart,
		      int Xend, int Yend)
{
   float max, min, mean1, mean2, var1, var2;
   float low_thresh, score, low_score = 0.0;
   float Threshold;
   int loop, count1, count2, x, y;

   /* Find max and min */
   max = min = Data[0][0];
   for (y = Ystart; y < Yend; y++)
      for (x = Xstart; x < Xend; x++)
	 if (Data[y][x] > max)
	    max = Data[y][x];
	 else if (Data[y][x] < min)
	    min = Data[y][x];

   /* Search for optimal threshold */
   for (loop = 1; loop < 1000; loop++)
   {

      /* Calculate means and variance of two classes */
      Threshold = min + loop * (max - min) / 1000;
      var1 = var2 = 0.0;
      mean1 = mean2 = 0.0;
      count1 = count2 = 1;

      for (y = Ystart; y < Yend; y++)
	 for (x = Xstart; x < Xend; x++)
	    if (Data[y][x] < Threshold)
	    {
	       count1++;
	       mean1 += Data[y][x];
	       var1 += Data[y][x] * Data[y][x];
	    }
	    else
	    {
	       count2++;
	       mean2 += Data[y][x];
	       var2 += Data[y][x] * Data[y][x];
	    }

      mean1 = mean1 / count1;
      mean2 = mean2 / count2;
      var1 = var1 / count1 - mean1 * mean1;
      var2 = var2 / count2 - mean2 * mean2;

      /* See if this is the best score */
      score = (count1 * var1 + count2 * var2)
	 / ((Yend - Ystart) * (Xend - Xstart));
      if ((loop == 1) || (score < low_score))
      {
	 low_score = score;
	 low_thresh = Threshold;
      }
   }

   /* Set optimal threshold */
   Threshold = low_thresh;

   return (Threshold);
}


/*---------------------------------------------------------------------------*/
/* Purpose: This routine performs thresholding of the entire image based on  */
/*          and using linear interpolation.                                  */
/*---------------------------------------------------------------------------*/
void LinearInterpolation(FLOAT_TYPE ** Data1, SHORT_TYPE ** Data2,
			 int Xdim, int Ydim)
{
   IM_TYPE *Image;
   FLOAT_TYPE **Data;
   int i, j, StepX, StepY, x, y, dx, dy, rx, ry, endX, endY;
   float Thresholds[100][100], AveThreshold = 0.0, StepThreshold;
   float LeftThreshold, RightThreshold, TopThreshold, BottomThreshold;
   char FileName[255];
   FILE *Fd;

   /* Create an output image to store debug information */
   if (Debug == TRUE)
   {
      fprintf(stderr, "The image 'debug-info' is now being created...\n\n");
      system("rm -f debug-info");
      Image = im_create("debug-info", FLOAT, Xdim, Ydim, 1);
      Data = (FLOAT_TYPE **) im_alloc2D(Image, FLOAT);
   }

   /* Determine region size */
   dx = Xdim / RegionX;
   dy = Ydim / RegionY;

   /* Determine the remainder of region size determination */
   rx = Xdim - dx * RegionX;
   ry = Ydim - dy * RegionY;

   /* Adjust the remainders */
   if (dx % 2 != 0)
      rx++;
   if (dy % 2 != 0)
      ry++;

   j = 0;
   /* Threshold each partition of the image */
   for (y = 0; y + dy <= Ydim; y += dy)
   {
      i = 0;
      for (x = 0; x + dx <= Xdim; x += dx)
      {
	 /* Determine the last column of a partition */
	 if (x + 2 * dx > Xdim)
	    endX = Xdim;
	 else
	    endX = x + dx;

	 /* Determine the last row of a partition */
	 if (y + 2 * dy > Ydim)
	    endY = Ydim;
	 else
	    endY = y + dy;

	 if (StatsThreshold == TRUE)
	    Thresholds[j][i] = StatThresholding(Data1, x, y, endX, endY);
	 else if (EdgeThreshold == TRUE)
	    Thresholds[j][i] = EdgeThresholding(Data1, x, y, endX, endY);
	 else if (OptimalThreshold == TRUE)
	    Thresholds[j][i] = OptThresholding(Data1, x, y, endX, endY);
	 else if (StaticThreshold == TRUE)
	    Thresholds[j][i] = UserThreshold;

	 /* Print threshold in each region */
	 if (Debug == TRUE)
	 {
	    if (Thresholds[j][i] >= 0)
	       printf("  %5.2f ", Thresholds[j][i]);
	    else
	       printf(" -%5.2f ", Thresholds[j][i] * -1.0);

	    if (x + (2 * dx) > Xdim)
	       printf("\n\n");
	 }
	 AveThreshold += Thresholds[j][i];
	 i++;
      }
      j++;
   }

   if (Debug == TRUE)
   {
      printf("--------\n");
      printf(" %5.2f\n\n", AveThreshold / (i * j));
   }

   /* Write threshold value to temporary file */
   if (StoreThreshold == TRUE)
   {
      strcat(FileName, "/tmp/");
      strcat(FileName, (char *) getenv("USER"));
      if ((Fd = fopen(FileName, "w")) == NULL)
	 Error("Could not open output file");
      fprintf(Fd, "%d\n", (int) AveThreshold / (i * j));
      fclose(Fd);
   }

   /* Threshold the image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 /* Left top corner */
	 if (y < dy / 2 && x < dx / 2)
	 {
	    if (Data1[y][x] < Thresholds[0][0])
	       Data2[y][x] = 0;
	    else
	       Data2[y][x] = 1;

	    if (Debug == TRUE)
	       Data[y][x] = Thresholds[0][0];
	 }

	 /* Linear interpolate between top left and right corner */
	 if (y < dy / 2 && x >= dx / 2 && x < Xdim - (dx / 2) - rx)
	 {
	    i = (x - (dx / 2)) / dx;
	    StepX = x - (dx / 2) - (dx * i) + 1;
	    LeftThreshold = Thresholds[0][i];
	    RightThreshold = Thresholds[0][i + 1];
	    StepThreshold = LeftThreshold + StepX *
	       ((RightThreshold - LeftThreshold) / dx);

	    if (Data1[y][x] < StepThreshold)
	       Data2[y][x] = 0;
	    else
	       Data2[y][x] = 1;

	    if (Debug == TRUE)
	       Data[y][x] = StepThreshold;
	 }

	 /* Right top corner */
	 if (y < dy / 2 && x >= Xdim - (dx / 2) - rx)
	 {
	    if (Data1[y][x] < Thresholds[0][RegionX - 1])
	       Data2[y][x] = 0;
	    else
	       Data2[y][x] = 1;

	    if (Debug == TRUE)
	       Data[y][x] = Thresholds[0][RegionX - 1];
	 }

	 /* Linear interpolate between top and bottom left corner */
	 if ((x < dx / 2) && (y >= dy / 2) && (y < Ydim - (dy / 2) - ry))
	 {
	    i = (y - (dy / 2)) / dy;
	    StepY = y - (dy / 2) - (dy * i) + 1;
	    TopThreshold = Thresholds[i][0];
	    BottomThreshold = Thresholds[i + 1][0];
	    StepThreshold = TopThreshold + StepY *
	       ((BottomThreshold - TopThreshold) / dy);

	    if (Data1[y][x] < StepThreshold)
	       Data2[y][x] = 0;
	    else
	       Data2[y][x] = 1;
	    if (Debug == TRUE)
	       Data[y][x] = StepThreshold;
	 }

	 /* BiLinear interpolate the center */
	 if ((x >= dx / 2) && (x < Xdim - (dx / 2) - rx) &&
	     (y >= dy / 2) && (y < Ydim - (dy / 2) - ry))
	 {
	    i = (x - (dx / 2)) / dx;
	    j = (y - (dy / 2)) / dy;
	    StepX = x - (dx / 2) - (dx * i) + 1;
	    StepY = y - (dy / 2) - (dy * j) + 1;
	    TopThreshold = Thresholds[j][i] + StepX *
	       ((Thresholds[j][i + 1] - Thresholds[j][i]) / dx);
	    BottomThreshold = Thresholds[j + 1][i] + StepX *
	       ((Thresholds[j + 1][i + 1] - Thresholds[j + 1][i]) / dx);
	    StepThreshold = TopThreshold + StepY *
	       ((BottomThreshold - TopThreshold) / dy);

	    if (Data1[y][x] < StepThreshold)
	       Data2[y][x] = 0;
	    else
	       Data2[y][x] = 1;

	    if (Debug == TRUE)
	       Data[y][x] = StepThreshold;
	 }

	 /* Linear interpolate between top and bottom right corner */
	 if ((x >= Xdim - (dx / 2) - rx) && (y >= dy / 2) &&
	     (y < Ydim - (dy / 2) - ry))
	 {
	    i = (y - (dy / 2)) / dy;
	    StepY = y - (dy / 2) - (dy * i) + 1;
	    TopThreshold = Thresholds[i][RegionX - 1];
	    BottomThreshold = Thresholds[i + 1][RegionX - 1];
	    StepThreshold = TopThreshold + StepY *
	       ((BottomThreshold - TopThreshold) / dy);

	    if (Data1[y][x] < StepThreshold)
	       Data2[y][x] = 0;
	    else
	       Data2[y][x] = 1;

	    if (Debug == TRUE)
	       Data[y][x] = StepThreshold;
	 }

	 /* Left bottom corner */
	 if (y >= Ydim - (dy / 2) - ry && x < dx / 2)
	 {
	    if (Data1[y][x] < Thresholds[RegionY - 1][0])
	       Data2[y][x] = 0;
	    else
	       Data2[y][x] = 1;

	    if (Debug == TRUE)
	       Data[y][x] = Thresholds[RegionY - 1][0];

	 }

	 /* Linear interpolate between bottom left and right corner */
	 if ((y >= Ydim - (dy / 2) - ry) && (x >= dx / 2) &&
	     (x < Xdim - (dx / 2) - rx))
	 {
	    i = (x - (dx / 2)) / dx;
	    StepX = x - (dx / 2) - (dx * i) + 1;
	    LeftThreshold = Thresholds[RegionY - 1][i];
	    RightThreshold = Thresholds[RegionY - 1][i + 1];
	    StepThreshold = LeftThreshold + StepX *
	       ((RightThreshold - LeftThreshold) / dx);

	    if (Data1[y][x] < StepThreshold)
	       Data2[y][x] = 0;
	    else
	       Data2[y][x] = 1;

	    if (Debug == TRUE)
	       Data[y][x] = StepThreshold;
	 }

	 /* Right bottom corner */
	 if (y >= Ydim - (dy / 2) - ry && x >= Xdim - (dx / 2) - rx)
	 {
	    if (Data1[y][x] < Thresholds[RegionY - 1][RegionX - 1])
	       Data2[y][x] = 0;
	    else
	       Data2[y][x] = 1;

	    if (Debug == TRUE)
	       Data[y][x] = Thresholds[RegionY - 1][RegionX - 1];
	 }
      }

   /* Write debug information to output image */
   if (Debug == TRUE)
   {
      im_write(Image, FLOAT, (char *) &(Data[0][0]));
      im_free2D((char **) Data);
   }
}


/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];

   IM_TYPE *Image1;
   IM_TYPE *Image2;
   FLOAT_TYPE **Data1;
   SHORT_TYPE **Data2;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("THRESH Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      case 's':
	 StatsThreshold = TRUE;
	 break;
      case 'e':
	 EdgeThreshold = TRUE;
	 break;
      case 'o':
	 OptimalThreshold = TRUE;
	 break;
      case 't':
	 if (sscanf(argv[++i], "%f", &UserThreshold) == 0)
	    Error("Could not get Threshold argument");
	 StaticThreshold = TRUE;
	 break;
      case 'r':
	 if (sscanf(argv[++i], "%d ", &RegionX) == 0)
	    Error("Could not get RegionX argument");
	 if (sscanf(argv[++i], "%d ", &RegionY) == 0)
	    Error("Could not get RegionY argument");
	 if (RegionX <= 0 || RegionX > 50 || RegionY <= 0 || RegionY > 50)
	    Error("Invalid number encountered (range 1-50)");
	 break;
      case 'g':
	 StoreThreshold = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: thresh [options] infile outfile\n");
      fprintf(stderr, "       [-d]     Print debugging information\n");
      fprintf(stderr, "       [-s]     Statistics based threshold\n");
      fprintf(stderr, "       [-e]     Edge strength based threshold\n");
      fprintf(stderr, "       [-o]     Optimal threshold\n");
      fprintf(stderr, "       [-t #]   Static Threshold (0)\n");
      fprintf(stderr, "       [-r # #] Number of partitions in X and Y (1,1)\n");
      exit(1);
   }

   /* Check if thresholding method specified */
   if (StatsThreshold == FALSE && EdgeThreshold == FALSE &&
       OptimalThreshold == FALSE && StaticThreshold == FALSE)
      Error("NO thresholding method option (s, e, o, t) encountered");

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (PixType == COLOR || PixType == PSEUDO || PixType == JPEG_RGB || PixType == COMPLEX)
      Error("Can not threshold color or complex images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, SHORT, Xdim, Ydim, Zdim);
   Data2 = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);

   /* Perform thresholding */
   LinearInterpolation(Data1, Data2, Xdim, Ydim);

   /* Write information to output image */
   im_write(Image2, SHORT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);

   return (0);
}
