/*--------------------------------------------------------------------------*/
/* Program:  to_color.c                                                     */
/*                                                                          */
/* Purpose:  This program converts 3D multichannel images to RGB color      */
/*           images in KUIM format.  The type of color conversion needed    */
/*           is specified by input switches.                                */
/*                                                                          */
/* Author:   John Gauch, Xiaodong Hong                                      */
/*                                                                          */
/* Date:     September 28, 1996 - Original program.                         */
/*           November 1 1996    - Add uvw xyz color systems.                */
/*           April 8, 1997      - Add HIS, U*V*W* and Lab color systems.    */
/*                                                                          */
/* Note:     Copyright (C) The University of Kansas, 1994                   */
/*--------------------------------------------------------------------------*/
#include <IM.h>
#define CONV_RGB  1
#define CONV_UVW  2
#define CONV_YIQ  3
#define CONV_XYZ  4
#define CONV_UVW2 5
#define CONV_HIS  6
#define CONV_LAB  7
static float X_0 = 0.982;
static float Y_0 = 1.00;
static float Z_0 = 1.183;
static float u_0 = 0.201;
static float v_0 = 0.307;

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   FLOAT_TYPE ***Data1;
   COLOR_TYPE **Data2;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int Convert = CONV_RGB;
   int i = 0, j, k;
   float X, Y, Z, U, V, W, I, Q, R, G, B, U2, V2, W2, temp;
   float H, S, L, a2, b2, u, v;

   /* Interpret program options */
   printf("TO_COLOR Program - KUIM Version 3.0\n\n");

   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'r':
	 Convert = CONV_RGB;
	 break;
      case 'y':
	 Convert = CONV_YIQ;
	 break;
      case 'u':
	 Convert = CONV_UVW;
	 break;
      case 'x':
	 Convert = CONV_XYZ;
	 break;
      case 'U':
	 Convert = CONV_UVW2;
	 break;
      case 'h':
	 Convert = CONV_HIS;
	 break;
      case 'l':
	 Convert = CONV_LAB;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: to_color [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-rgb]  Input is RGB format\n");
      fprintf(stderr, "       [-uvw]  Input is UVW format\n");
      fprintf(stderr, "       [-yiq]  Input is YIQ format\n");
      fprintf(stderr, "       [-xyz]  Input is XYZ format\n");
      fprintf(stderr, "       [-UVW]  Input is U*V*W* format\n");
      fprintf(stderr, "       [-his]  Input is HIS format\n");
      fprintf(stderr, "       [-lab]  Input is Lab format\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 3)
      Error("Can only process 3D images");
   Data1 = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0][0]));

   /* Create output image */
   Image2 = im_create(Name2, COLOR, Xdim, Ydim, 1);
   Data2 = (COLOR_TYPE **) im_alloc2D(Image2, COLOR);

   /* Manipulate input image data */
   for (k = 0; k < Ydim; k++)
      for (j = 0; j < Xdim; j++)
      {
	 if (Convert == CONV_RGB)
	 {
	    Data2[k][j].r = (unsigned char) Data1[0][k][j];
	    Data2[k][j].g = (unsigned char) Data1[1][k][j];
	    Data2[k][j].b = (unsigned char) Data1[2][k][j];
	 }

	 else if (Convert == CONV_UVW)
	 {
	    U = Data1[0][k][j];
	    V = Data1[1][k][j];
	    W = Data1[2][k][j];
	    R = 2.431 * U + 0.331 * V - 0.576 * W;
	    G = -1.518 * U + 2.084 * V - 0.057 * W;
	    B = 1.440 * U - 2.825 * V + 1.803 * W;
	    if (R < 0)
	       R = 0;
	    if (R > 255)
	       R = 255;
	    if (G < 0)
	       G = 0;
	    if (G > 255)
	       G = 255;
	    if (B < 0)
	       B = 0;
	    if (B > 255)
	       B = 255;
	    Data2[k][j].r = (unsigned char) R;
	    Data2[k][j].g = (unsigned char) G;
	    Data2[k][j].b = (unsigned char) B;
	 }

	 else if (Convert == CONV_YIQ)
	 {
	    Y = Data1[0][k][j];
	    I = Data1[1][k][j];
	    Q = Data1[2][k][j];
	    R = 1.000 * Y + 0.956 * I + 0.621 * Q;
	    G = 1.000 * Y - 0.273 * I - 0.647 * Q;
	    B = 1.000 * Y - 1.104 * I + 1.701 * Q;
	    if (R < 0)
	       R = 0;
	    if (R > 255)
	       R = 255;
	    if (G < 0)
	       G = 0;
	    if (G > 255)
	       G = 255;
	    if (B < 0)
	       B = 0;
	    if (B > 255)
	       B = 255;
	    Data2[k][j].r = (unsigned char) R;
	    Data2[k][j].g = (unsigned char) G;
	    Data2[k][j].b = (unsigned char) B;
	 }

	 else if (Convert == CONV_XYZ)
	 {
	    X = Data1[0][k][j];
	    Y = Data1[1][k][j];
	    Z = Data1[2][k][j];
	    R = 2.364 * X - 0.896 * Y - 0.468 * Z;
	    G = -0.515 * X + 1.425 * Y + 0.088 * Z;
	    B = 0.005 * X - 0.014 * Y + 1.009 * Z;
	    if (R < 0)
	       R = 0;
	    if (R > 255)
	       R = 255;
	    if (G < 0)
	       G = 0;
	    if (G > 255)
	       G = 255;
	    if (B < 0)
	       B = 0;
	    if (B > 255)
	       B = 255;
	    Data2[k][j].r = (unsigned char) R;
	    Data2[k][j].g = (unsigned char) G;
	    Data2[k][j].b = (unsigned char) B;
	 }

	 else if (Convert == CONV_UVW2)
	 {
	    U2 = Data1[0][k][j];
	    V2 = Data1[1][k][j];
	    W2 = Data1[2][k][j];
	    temp = (W2 + 17.0) / 25.0;
	    Y = temp * temp * temp / 100.0;
	    u = U2 / (13.0 * W2) + u_0;
	    v = V2 / (13.0 * W2) + v_0;
	    if (v == 0)
	       X = 0;
	    else
	       X = 6.0 * u * Y / (4.0 * v);
	    if (u == 0)
	       Z = 0;
	    else
	       Z = (4.0 * X - u * X - 15.0 * Y * u) / (3 * u);
	    R = 2.364 * X - 0.896 * Y - 0.468 * Z;
	    G = -0.515 * X + 1.425 * Y + 0.088 * Z;
	    B = 0.005 * X - 0.014 * Y + 1.009 * Z;
	    if (R < 0)
	       R = 0;
	    if (R > 255)
	       R = 255;
	    if (G < 0)
	       G = 0;
	    if (G > 255)
	       G = 255;
	    if (B < 0)
	       B = 0;
	    if (B > 255)
	       B = 255;
	    Data2[k][j].r = (unsigned char) R;
	    Data2[k][j].g = (unsigned char) G;
	    Data2[k][j].b = (unsigned char) B;
	 }

	 else if (Convert == CONV_HIS)
	 {
	    H = Data1[0][k][j];
	    W2 = Data1[1][k][j];
	    S = Data1[2][k][j];
	    /* Problem with the sign of cos(H)? */
	    u = S / (13.0 * W2) * (float) cos((double) H) + u_0;
	    v = v_0 + (u - u_0) * (float) tan((double) H);
	    temp = (W2 + 17.0) / 25.0;
	    Y = temp * temp * temp / 100.0;
	    if (v == 0)
	       X = 0;
	    else
	       X = 6.0 * u * Y / (4.0 * v);
	    if (u == 0)
	       Z = 0;
	    else
	       Z = (4.0 * X - u * X - 15.0 * Y * u) / (3 * u);
	    R = 2.364 * X - 0.896 * Y - 0.468 * Z;
	    G = -0.515 * X + 1.425 * Y + 0.088 * Z;
	    B = 0.005 * X - 0.014 * Y + 1.009 * Z;
	    if (R < 0)
	       R = 0;
	    if (R > 255)
	       R = 255;
	    if (G < 0)
	       G = 0;
	    if (G > 255)
	       G = 255;
	    if (B < 0)
	       B = 0;
	    if (B > 255)
	       B = 255;
	    Data2[k][j].r = (unsigned char) R;
	    Data2[k][j].g = (unsigned char) G;
	    Data2[k][j].b = (unsigned char) B;

	 }

	 else if (Convert == CONV_LAB)
	 {
	    L = Data1[0][k][j];
	    a2 = Data1[1][k][j];
	    b2 = Data1[2][k][j];
	    temp = (L / 25.0);
	    Y = temp * temp * temp * Y_0 / 100.0;
	    temp = a2 / 500.0 + (float) pow((double) (Y / Y_0), 1.0 / 3.0);
	    X = temp * temp * temp * X_0;
	    temp = (float) pow((double) (Y / Y_0), 1.0 / 3.0) - b2 / 200.0;
	    Z = temp * temp * temp * Z_0;
	    R = 2.364 * X - 0.896 * Y - 0.468 * Z;
	    G = -0.515 * X + 1.425 * Y + 0.088 * Z;
	    B = 0.005 * X - 0.014 * Y + 1.009 * Z;
	    if (R < 0)
	       R = 0;
	    if (R > 255)
	       R = 255;
	    if (G < 0)
	       G = 0;
	    if (G > 255)
	       G = 255;
	    if (B < 0)
	       B = 0;
	    if (B > 255)
	       B = 255;
	    Data2[k][j].r = (unsigned char) R;
	    Data2[k][j].g = (unsigned char) G;
	    Data2[k][j].b = (unsigned char) B;
	 }
      }

   /* Write information to output image */
   im_write(Image2, COLOR, (char *) &(Data2[0][0]));
   im_free3D((char ***) Data1);
   im_free2D((char **) Data2);
   return (0);
}
