/*---------------------------------------------------------------------------*/
/* Program:  zoom.c                                                          */
/*                                                                           */
/* Purpose:  This program zooms into an image using the specified zoom       */
/*           factor and center location.  Both nearest neighbor and          */
/*           bilinear interpolation are supported.                           */
/*                                                                           */
/* Author:   John Gauch, Edu Metz                                            */
/*                                                                           */
/* Date:     October 26, 1994                                                */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define NEAREST  1
#define BILINEAR 2

/* Image variables */
char Name1[50];
char Name2[50];
IM_TYPE *Image1;
IM_TYPE *Image2;
int PixType, Xdim, Ydim, Zdim, DimCnt;

float ShiftX, ShiftY;
float OldX, OldY;
float PartX, PartY;
int x, y;

int Method = BILINEAR;
int CenterX = -1;
int CenterY = -1;
float Zoom = 2.0;

/*---------------------------------------------------------------------------*/
/* Purpose: Zooms into an Grey image.                                        */
/*---------------------------------------------------------------------------*/
void GreyZoom()
{
   FLOAT_TYPE **Data1;
   FLOAT_TYPE **Data2;

   /* Read input image */
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, FLOAT, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);

   /* Calculate some zooming constants */
   if (CenterX < 0)
      CenterX = Xdim / 2;
   if (CenterY < 0)
      CenterY = Ydim / 2;
   if (Zoom <= 0)
      Zoom = 1.0;
   ShiftX = CenterX - Xdim / (2 * Zoom);
   ShiftY = CenterY - Ydim / (2 * Zoom);

   /* Calculate zoomed image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 /* Find pixel coordinates in input image */
	 OldY = y / Zoom + ShiftY;
	 OldX = x / Zoom + ShiftX;
	 if ((OldX > 0) && (OldY > 0) && (OldX < Xdim - 1) && (OldY < Ydim - 1))
	 {
	    /* Nearest neighbor sampling */
	    if (Method == NEAREST)
	       Data2[y][x] = Data1[(int) (OldY + 0.5)][(int) (OldX + 0.5)];

	    /* Bilinear interpolation sampling */
	    else
	    {
	       PartX = OldX - (int) OldX;
	       PartY = OldY - (int) OldY;
	       Data2[y][x] = (1 - PartX) * (1 - PartY) * Data1[(int) OldY][(int) OldX] +
		  (1 - PartX) * (PartY) * Data1[(int) OldY + 1][(int) OldX] +
		  (PartX) * (1 - PartY) * Data1[(int) OldY][(int) OldX + 1] +
		  (PartX) * (PartY) * Data1[(int) OldY + 1][(int) OldX + 1];
	    }
	 }
	 /* Assume zeros outside image */
	 else
	    Data2[y][x] = 0.0;
      }

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data2[0][0]));

   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose: Zooms into an Color image.                                       */
/*---------------------------------------------------------------------------*/
void ColorZoom()
{
   COLOR_TYPE **Data1;
   COLOR_TYPE **Data2;

   /* Read input image */
   Data1 = (COLOR_TYPE **) im_alloc2D(Image1, COLOR);
   im_read(Image1, COLOR, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, COLOR, Xdim, Ydim, Zdim);
   Data2 = (COLOR_TYPE **) im_alloc2D(Image2, COLOR);

   /* Calculate some zooming constants */
   if (CenterX < 0)
      CenterX = Xdim / 2;
   if (CenterY < 0)
      CenterY = Ydim / 2;
   if (Zoom <= 0)
      Zoom = 1.0;
   ShiftX = CenterX - Xdim / (2 * Zoom);
   ShiftY = CenterY - Ydim / (2 * Zoom);

   /* Calculate zoomed image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 /* Find pixel coordinates in input image */
	 OldY = y / Zoom + ShiftY;
	 OldX = x / Zoom + ShiftX;
	 if ((OldX > 0) && (OldY > 0) && (OldX < Xdim - 1) && (OldY < Ydim - 1))
	 {
	    /* Nearest neighbor sampling */
	    if (Method == NEAREST)
	    {
	       Data2[y][x].r = Data1[(int) (OldY + 0.5)][(int) (OldX + 0.5)].r;
	       Data2[y][x].g = Data1[(int) (OldY + 0.5)][(int) (OldX + 0.5)].g;
	       Data2[y][x].b = Data1[(int) (OldY + 0.5)][(int) (OldX + 0.5)].b;
	    }
	    /* Bilinear interpolation sampling */
	    else
	    {
	       PartX = OldX - (int) OldX;
	       PartY = OldY - (int) OldY;
	       Data2[y][x].r = (unsigned char)
		  ((1 - PartX) * (1 - PartY) * Data1[(int) OldY][(int) OldX].r +
	       (1 - PartX) * (PartY) * Data1[(int) OldY + 1][(int) OldX].r +
	       (PartX) * (1 - PartY) * Data1[(int) OldY][(int) OldX + 1].r +
	       (PartX) * (PartY) * Data1[(int) OldY + 1][(int) OldX + 1].r);

	       Data2[y][x].g = (unsigned char)
		  ((1 - PartX) * (1 - PartY) * Data1[(int) OldY][(int) OldX].g +
	       (1 - PartX) * (PartY) * Data1[(int) OldY + 1][(int) OldX].g +
	       (PartX) * (1 - PartY) * Data1[(int) OldY][(int) OldX + 1].g +
	       (PartX) * (PartY) * Data1[(int) OldY + 1][(int) OldX + 1].g);

	       Data2[y][x].b = (unsigned char)
		  ((1 - PartX) * (1 - PartY) * Data1[(int) OldY][(int) OldX].b +
	       (1 - PartX) * (PartY) * Data1[(int) OldY + 1][(int) OldX].b +
	       (PartX) * (1 - PartY) * Data1[(int) OldY][(int) OldX + 1].b +
	       (PartX) * (PartY) * Data1[(int) OldY + 1][(int) OldX + 1].b);
	    }
	 }
	 /* Assume zeros outside image */
	 else
	    Data2[y][x].r = Data2[y][x].g = Data2[y][x].b = 0;
      }

   /* Write information to output image */
   im_write(Image2, COLOR, (char *) &(Data2[0][0]));

   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
/* Program variables */
   int Debug = FALSE;
   int i = 0;

   /* Interpret program options */
   printf("ZOOM Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'b':
	 Method = BILINEAR;
	 break;
      case 'n':
	 Method = NEAREST;
	 break;
      case 'z':
	 if (sscanf(argv[++i], "%f", &Zoom) == 0)
	    Error("Could not get zoom factor");
	 break;
      case 'c':
	 if (sscanf(argv[++i], "%d", &CenterX) == 0)
	    Error("Could not get center x");
	 if (sscanf(argv[++i], "%d", &CenterY) == 0)
	    Error("Could not get center y");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: zoom [options] infile outfile\n");
      fprintf(stderr, "       [-b]      Bilinear sampling\n");
      fprintf(stderr, "       [-n]      Nearest neighbor sampling\n");
      fprintf(stderr, "       [-z #]    Zoom factor (2,0)\n");
      fprintf(stderr, "       [-c # #]  Zoom center (Xdim/2, Ydim/2)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");

   if ((PixType == COLOR) || (PixType == PSEUDO) || PixType == JPEG_RGB)
      ColorZoom();
   else
      GreyZoom();
   return (0);
}
